"""
File: fasta.py
Description: Instantiate a FASTA file object.
CreateDate: 2021/11/26
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper, StringIO
from typing import Union, Any, Generator, Tuple
from itertools import groupby
from collections import Counter, defaultdict
from os.path import abspath
from gzip import GzipFile
from re import findall
from pandas import DataFrame
from click import echo
from pybioinformatic.util import FuncDict
from pybioinformatic.sequence import Nucleotide, Protein


class Fasta:
    def __init__(self, path: Union[str, TextIOWrapper]):
        """
        Initialize name, __open, and seq_num attributions.
        """
        if isinstance(path, str):
            self.name = abspath(path)
            if path.endswith('gz'):
                self.__open = GzipFile(path)
                self.seq_num = sum(1 for line in self.__open if str(line, 'utf8').startswith('>'))
                self.__open.seek(0)
            else:
                self.__open = open(path)
                self.seq_num = sum(1 for line in open(path) if line.startswith('>'))
        elif isinstance(path, TextIOWrapper):
            self.name = abspath(path.name)
            if path.name == '<stdin>' or '/dev' in path.name:
                string_io = StringIO()
                for line in path:
                    string_io.write(line)
                string_io.seek(0)
                self.__open = string_io
                self.seq_num = sum(1 for line in self.__open if line.startswith('>'))
                string_io.seek(0)
            else:
                if path.name.endswith('gz'):
                    self.__open = GzipFile(path.name)
                    self.seq_num = sum(1 for line in self.__open if str(line, 'utf8').startswith('>'))
                    self.__open.seek(0)
                else:
                    self.__open = path
                    self.seq_num = sum(1 for line in self.__open if line.startswith('>'))
                    self.__open.seek(0)
        else:
            echo(f'\033[31mInvalid type: {type(path)}.\033[0m', err=True)
            exit()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def __seek_zero(self):
        try:
            self.__open.seek(0)
        except Exception:
            pass

    def parse(self, parse_id: bool = True) -> Generator[Union[Protein, Nucleotide], Any, None]:
        """A FASTA file generator that returns one Nucleotide or Protein object at one time."""
        if self.name.endswith('gz'):
            fa_generator = (ret[1] for ret in groupby(self.__open, lambda line: str(line, 'utf8').startswith('>')))
        else:
            fa_generator = (ret[1] for ret in groupby(self.__open, lambda line: line.startswith('>')))
        self.__seek_zero()
        for g in fa_generator:
            if self.name.endswith('gz'):
                seq_id = str(g.__next__(), 'utf8').strip()
                seq = ''.join(str(line, 'utf8').strip() for line in fa_generator.__next__())
            else:
                seq_id = g.__next__().strip()
                seq = ''.join(line.strip() for line in fa_generator.__next__())
            if parse_id:
                if '\t' in seq_id:
                    seq_id = seq_id.split('\t')[0]
                elif '|' in seq_id:
                    seq_id = seq_id.split('|')[0]
                else:
                    seq_id = seq_id.split(' ')[0]
            if set(seq.upper()) - {'A', 'G', 'C', 'T', 'N'}:
                yield Protein(seq_id, seq)
            else:
                yield Nucleotide(seq_id, seq)

    def to_dict(self, parse_id: bool = False) -> dict:
        """Parse fasta as dict."""
        seq_dict = {}
        for nucl_obj in self.parse(parse_id):
            if nucl_obj.id not in seq_dict:
                seq_dict[nucl_obj.id] = nucl_obj.seq
            else:
                echo(f'\033[31mError: FASTA file has repeat id {nucl_obj.id}.', err=True)
                self.__seek_zero()
                exit()
        self.__seek_zero()
        return seq_dict

    def to_dataframe(self, parse_id: bool = True) -> DataFrame:
        """Parse fasta as pandas.DataFrame"""
        seq_dict = self.to_dict(parse_id)
        for k, v in seq_dict.items():
            seq_dict[k] = list(v)
        return DataFrame(seq_dict)

# File format conversion method=========================================================================================
    def merge_sequence(self, parse_id: bool = False) -> Generator[Union[Protein, Nucleotide], Any, None]:
        """Make each sequence to be displayed on a single line."""
        for seq_obj in self.parse(parse_id):
            yield seq_obj

    def split_sequence(self, parse_id: bool = False, char_num: int = 60) -> Generator[Union[Protein, Nucleotide], Any, None]:
        """Make each sequence to be displayed in multiple lines."""
        for seq_obj in self.parse(parse_id):
            seq_obj = seq_obj.display_set(char_num)
            yield seq_obj

    def fa2tab(self, parse_id: bool = False) -> Generator[str, Any, None]:
        """Convert fasta to tab delimited txt text files."""
        for seq_obj in self.parse(parse_id):
            yield f'{seq_obj.id}\t{seq_obj.seq}'

# Other method==========================================================================================================
    def get_size(self) -> int:
        size = sum(len(seq_obj) for seq_obj in self.parse())
        return size

    def get_longest_seq(
        self,
        regular_exp: str = r'\w+.\w+',
        inplace_id: bool = False
    ) -> Generator[Union[Protein, Nucleotide], Any, None]:
        """Get the longest transcript of each gene locus."""
        all_seq_dict = self.to_dict(False)  # {seq_id: seq}
        longest_seq_dict = {}  # {locus_id: seq}
        id_map_dict = {}  # {'Potri.001G000100': 'Potri.001G000100.3', 'Potri.001G000200': 'Potri.001G000200.1', ...}
        if inplace_id:
            for seq_obj in self.parse(False):
                gene_id = findall(regular_exp, seq_obj.id)[0]
                if gene_id not in id_map_dict:
                    id_map_dict[gene_id] = seq_obj.id
                else:
                    if len(seq_obj) >= len(all_seq_dict[id_map_dict[gene_id]]):
                        id_map_dict[gene_id] = seq_obj.id
            for locus, longest_seq_id in id_map_dict.items():
                longest_seq_dict[locus] = all_seq_dict[longest_seq_id]
        else:
            for seq_obj in self.parse(False):
                gene_id = findall(regular_exp, seq_obj.id)[0]
                if gene_id not in id_map_dict:
                    id_map_dict[gene_id] = seq_obj.id
                else:
                    if seq_obj.len >= len(all_seq_dict[id_map_dict[gene_id]]):
                        id_map_dict[gene_id] = seq_obj.id
            for locus, longest_seq_id in id_map_dict.items():
                longest_seq_dict[longest_seq_id] = all_seq_dict[longest_seq_id]
        for seq_id, seq in longest_seq_dict.items():
            yield Protein(seq_id, seq) if set(seq.upper()) - {'A', 'G', 'C', 'T', 'N'} else Nucleotide(seq_id, seq)

    def filter_n(self, max_num=1) -> Generator[Union[Protein, Nucleotide], Any, None]:
        for nucl_obj in self.parse():
            if nucl_obj.seq.upper().count('N') < max_num:
                yield nucl_obj
            else:
                echo(f'{nucl_obj.id} has been filtered out.', err=True)

    def get_k_mer(self, k: int = 21, parse_id: bool = True):
        """Get K-mer sequence for each sequence from fasta file."""
        for nucl in self.parse(parse_id):
            yield from nucl.k_mer(k)

    def stat_k_mer(self, k: int = 21) -> FuncDict:
        """Count the frequency of each k-mer."""
        k_mer_list = (i for i in self.get_k_mer(k=k))
        seq_list = [i.seq for i in k_mer_list]
        count_dict = FuncDict(Counter(Counter(seq_list).values()))
        count_dict.sort_by_keys()
        return count_dict

    def rename(self, map_dict: dict, parse_seq_id: bool = True, remain: bool = True):
        for nucl_obj in self.parse(parse_id=parse_seq_id):
            if nucl_obj.id in map_dict:
                new_id = map_dict[nucl_obj.id]
                nucl_obj.id = new_id
            elif nucl_obj.id not in map_dict and remain:
                pass
            else:
                break
            yield nucl_obj

    def hard_mask_to_soft_mask(self, non_masked) -> Generator[Tuple[str, str], None, None]:
        with Fasta(non_masked) as no_masked:
            seq_dict = no_masked.to_dict(True)

        mask_pos = defaultdict(list)
        for seq_obj in self.parse(True):
            for gap in seq_obj.locate_gap():
                gap_s, gap_e = gap
                if seq_obj.seq[gap_s - 1:gap_e] != seq_dict[seq_obj.id][gap_s - 1:gap_e]:
                    mask_pos[seq_obj.id].append((gap_s, gap_e))

        for ID, rep_pos in mask_pos.items():
            soft_mask_seq = []
            for i in rep_pos:
                s, e = i
                if i == rep_pos[0]:
                    seq = [seq_dict[ID][:s - 1], seq_dict[ID][s - 1:e].lower()]
                elif i == rep_pos[-1]:
                    j = rep_pos[rep_pos.index(i) - 1]
                    le = j[1]
                    seq = [seq_dict[ID][le:s - 1], seq_dict[ID][s - 1:e].lower(), seq_dict[ID][e:]]
                else:
                    j = rep_pos[rep_pos.index(i) - 1]
                    le = j[1]
                    seq = [seq_dict[ID][le:s - 1], seq_dict[ID][s - 1:e].lower()]
                soft_mask_seq.extend(seq)
            soft_mask_seq = ''.join(soft_mask_seq)
            yield ID, soft_mask_seq



"""
File: timer.py
Description: Instantiate a timer class object
CreateDate: 2021/12/31
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from datetime import datetime
from click import echo


class Timer:
    def __init__(self, start_message: str = None, command_content: str = None):
        self.start_message = start_message
        self.command_content = command_content

    def __call__(self, function):
        def wrapper(*args, **kwargs):
            start_time = datetime.now().replace(microsecond=0)
            if self.start_message:
                echo(f"\033[36m[{datetime.now().replace(microsecond=0)}] {self.start_message}\033[0m", err=True)
            else:
                msg = function.__name__.replace('_', ' ')
                echo(f"\033[36m[{datetime.now().replace(microsecond=0)}] Start run {msg}.\033[0m", err=True)
            if self.command_content:
                echo(f"\033[36m[{datetime.now().replace(microsecond=0)}] {self.command_content}\033[0m", err=True)
            value = function(*args, **kwargs)
            end_time = datetime.now().replace(microsecond=0)
            echo(f'\033[36m[{datetime.now().replace(microsecond=0)}] Finish in {end_time - start_time}.\033[0m', err=True)
            return value
        return wrapper



"""
File: bed.py
Description: Instantiate a BED file object.
CreateDate: 2021/12/4
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union
from io import TextIOWrapper
from os.path import abspath
from pandas import read_table, Series
from click import open_file
from pybioinformatic.sequence import Nucleotide
from pybioinformatic.fasta import Fasta


class Bed:
    def __init__(self, path: Union[str, TextIOWrapper]):
        self.header = ['Chromosome', 'Start', 'End', 'Name', 'Score', 'Strand', 'ThickStart', 'ThickEnd',
                       'ItemRGB', 'BlockCount', 'BlockSize', 'BlockStart']
        if isinstance(path, str):
            self.name = abspath(path)
            self.__open = open(path)
            self.line_num = sum(1 for _ in self.__open)
            self.__open.seek(0)
        else:
            self.name = abspath(path.name)
            if path.name == '<stdin>':
                self.__open = open_file('-').readlines()
                self.line_num = sum(1 for _ in self.__open)
            else:
                self.__open = path
                self.line_num = sum(1 for _ in self.__open)
                self.__open.seek(0)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def to_dataframe(self):
        df = read_table(self.__open, header=None, names=self.header)
        return df

    def get_bed_dict(self) -> dict:
        bed_dict = {}  # {Chr1: [{start: int, end: int, id: str, frame: str, strand: str}, ...], ...}
        for line in self.__open:
            if not line.startswith('#') and line.strip():
                split = line.strip().split('\t')
                item = {'start': int(split[1]), 'end': int(split[2]), 'id': split[3],
                        'frame': split[4], 'strand': split[5]}
                if split[0] in bed_dict:
                    bed_dict[split[0]].append(item)
                else:
                    bed_dict[split[0]] = [item]
        return bed_dict

# Interval arithmetic method============================================================================================
    def merge_interval_by_feature(self):
        def apply_func(row: Series):
            chr_name = row['Chromosome'].unique()[0]
            strand = row['Strand'].unique()[0]
            start, end = row['Start'].min(), row['End'].max()
            data = [chr_name, start, end, strand]
            index = ['Chromosome', 'Start', 'End', 'Strand']
            ser = Series(data=data, index=index)
            return ser

        df = self.to_dataframe()
        merge_bed = df.groupby(by='Name').apply(func=apply_func)
        return merge_bed

    def extend_interval(self, chr_len_dict: dict, upstream: int = 0, downstream: int = 0, both: int = None):
        def apply_func(row: Series):
            chr_name = row['Chromosome']
            start = row['Start']
            end = row['End']
            name = row['Name']
            score = row['Score']
            strand = row['Strand']
            max_length = chr_len_dict[chr_name]
            if both:
                start = start - both if start - both > 0 else 1
                end = end + both if end + both <= max_length else max_length
            else:
                if strand == '+':
                    start = start - upstream if start - upstream > 0 else 1
                    end = end + downstream if end + downstream <= max_length else max_length
                elif strand == '-':
                    start = start - downstream if start - downstream > 0 else 1
                    end = end + upstream if end + upstream <= max_length else max_length
            ser = Series(data=[chr_name, start, end, name, score, strand], index=self.header[:6])
            return ser

        df = self.to_dataframe()
        extend_bed = df.apply(func=apply_func, axis=1)
        return extend_bed

# Sequence extraction method============================================================================================
    @staticmethod
    def __judge_range(start: int, end: int, strand: str, up: int = 0, down: int = 0, both: int = 0) -> tuple:
        if both:
            start -= both
            end += both
            if start < 0:
                start = 0
            return start, end
        else:
            if strand == '+':
                start -= up
                end += down
                if start < 0:
                    start = 0
                return start, end
            else:
                start -= down
                end += up
                return start, end

    def extract_seq(self,
                    fasta_file: Union[str, TextIOWrapper],
                    use_id: bool = True,
                    up: int = 0,
                    down: int = 0,
                    both: int = 0,
                    extension: bool = True) -> Nucleotide:
        """
        Extract the sequence in the BED file from the reference sequence file
        :param fasta_file: Reference sequence FASTA file
        :param use_id: Use fourth column content in the BED file as sequence ID. Otherwise,
                       "chr_num:start-end(strand)" by default.
        :param up: Make start site of sequence to extent to upstream specified length.
        :param down: Make end site of sequence to extent to downstream specified length.
        :param both: Make sequence to extend specified length both end.
        :param extension: Specify whether the extended sequence includes the sequence itself
        :return: Nucleotide_obj_generator
        """
        # bed_dict = {Chr_num: [{start: int, end: int, id: str, frame: str, strand: str}]}
        bed_dict = self.get_bed_dict()
        for nucl_obj in Fasta(fasta_file).parse():
            if nucl_obj.id in bed_dict:
                seqs: list = bed_dict[nucl_obj.id]  # [{start: int, end: int, id: str, frame: str, strand: str}]
                for d in seqs:
                    new_start, new_end = self.__judge_range(d['start'], d['end'], d['strand'], up, down, both)
                    if extension:
                        sub_seq_obj = nucl_obj[new_start:new_end]
                        if d['strand'] == '-':
                            sub_seq_obj = -sub_seq_obj
                        if use_id:
                            sub_seq_obj.id = d['id']
                        else:
                            sub_seq_obj.id = f"{nucl_obj.id}:{new_start}-{new_end}({d['strand']})"
                        yield sub_seq_obj
                    else:
                        if both:
                            if d['strand'] == '+':
                                up_seq_obj = nucl_obj[new_start:d['start']]
                                down_seq_obj = nucl_obj[d['end']:new_end]
                                if use_id:
                                    up_seq_obj.id = f"{d['id']} upstream_{both}_nt"
                                    down_seq_obj.id = f"{d['id']} downstream_{both}_nt"
                                else:
                                    up_seq_obj.id = f"{nucl_obj.id}:{new_start}-{d['start']}({d['strand']})"
                                    down_seq_obj.id = f"{nucl_obj.id}:{d['end']}-{new_end}({d['strand']})"
                            else:
                                up_seq_obj = nucl_obj[d['end']:new_end]
                                up_seq_obj = -up_seq_obj
                                down_seq_obj = nucl_obj[new_start:d['start']]
                                down_seq_obj = -down_seq_obj
                                if use_id:
                                    up_seq_obj.id = f"{d['id']} upstream_{both}_nt"
                                    down_seq_obj.id = f"{d['id']} downstream_{both}_nt"
                                else:
                                    up_seq_obj.id = f"{nucl_obj.id}:{d['end']}-{new_end}({d['strand']})"
                                    down_seq_obj.id = f"{nucl_obj.id}:{new_start}-{d['start']}({d['strand']})"
                            yield up_seq_obj
                            yield down_seq_obj
                        else:
                            if up:
                                if d['strand'] == '+':
                                    up_seq_obj = nucl_obj[new_start:d['start']]
                                    if use_id:
                                        up_seq_obj.id = f"{d['id']} upstream_{up}_nt"
                                    else:
                                        up_seq_obj.id = f"{nucl_obj.id}:{new_start}-{d['start']}({d['strand']})"
                                else:
                                    up_seq_obj = nucl_obj[d['end']:new_end]
                                    up_seq_obj = -up_seq_obj
                                    if use_id:
                                        up_seq_obj.id = f"{d['id']} upstream_{up}_nt"
                                    else:
                                        up_seq_obj.id = f"{nucl_obj.id}:{d['end']}-{new_end}({d['strand']})"
                                yield up_seq_obj
                            if down:
                                if d['strand'] == '+':
                                    down_seq_obj = nucl_obj[d['end']:new_end]
                                    if use_id:
                                        down_seq_obj.id = f"{d['id']} downstream_{down}_nt"
                                    else:
                                        down_seq_obj.id = f"{nucl_obj.id}:{d['end']}-{new_end}({d['strand']})"
                                else:
                                    down_seq_obj = nucl_obj[new_start:d['start']]
                                    down_seq_obj = -down_seq_obj
                                    if use_id:
                                        down_seq_obj.id = f"{d['id']} downstream_{down}_nt"
                                    else:
                                        down_seq_obj.id = f"{nucl_obj.id}:{new_start}-{d['start']}({d['strand']})"
                                yield down_seq_obj



#!/usr/bin/env python
"""
File: fastq.py
Description: Instantiate a FASTA file object.
CreateDate: 2024/7/25
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper
from typing import Union, Any, Generator, Tuple
from os.path import abspath
import gzip
from random import randint, choice, choices
from re import sub
from tqdm import tqdm
from click import echo, open_file
from pybioinformatic.sequence import Sequence, Reads
from pybioinformatic.util import FuncDict
from pybioinformatic.task_manager import TaskManager


def _sample(fastq_file_path: str, flags):
    out_file = sub(
        pattern='.fastq$',
        repl='',
        string=sub(
            pattern='.fq$',
            repl='',
            string=sub(
                pattern='.gz$',
                repl='',
                string=fastq_file_path
            )
        )
    ) + f'_sample{len(flags)}.fastq.gz'
    i = 0
    with Fastq(fastq_file_path) as fastq_obj, gzip.open(out_file, 'wb') as o:
        for record in fastq_obj.parse():
            if i > max(flags):
                break
            if i in flags:
                o.write(str(record).encode('utf-8'))
            i += 1


class Fastq:
    def __seek_zero(self):
        try:
            self.__open.seek(0)
        except AttributeError:
            pass

    def __check_integrity(self):
        if self.line_num % 4 != 0:
            echo(f'\033[31mError: Truncated fastq file "{self.name}".\033[0m', err=True)
            exit()
        else:
            self.num_reads = int(self.line_num / 4)

    def __init__(self, path: Union[str, TextIOWrapper]):
        """
        Initialize name, __open, num_line and num_reads attributions.
        """
        if isinstance(path, str):
            self.name = abspath(path)
            if path.endswith('gz'):
                self.__open = gzip.GzipFile(path)
            else:
                self.__open = open(path)
            self.line_num = sum(1 for _ in self.__open)
            self.__open.seek(0)
            self.__check_integrity()
        elif isinstance(path, TextIOWrapper):
            self.name = abspath(path.name)
            if path.name == '<stdin>':
                self.__open = open_file('-').readlines()
                self.line_num = len(self.__open)
                self.__check_integrity()
            else:
                if path.name.endswith('gz'):
                    self.__open = gzip.GzipFile(path.name)
                else:
                    self.__open = path
                self.line_num = sum(1 for _ in self.__open)
                self.__open.seek(0)
                self.__check_integrity()
        else:
            echo(f'\033[31mInvalid type: {type(path)}.\033[0m', err=True)
            exit()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def parse(self):
        while 1:
            record_id = self.__open.readline().strip()
            if not record_id:
                self.__seek_zero()
                break
            record_seq = self.__open.readline().strip()
            record_desc = self.__open.readline().strip()
            record_quality = self.__open.readline().strip()
            if self.name.endswith('gz'):
                record_id = str(record_id, 'utf8')
                record_seq = str(record_seq, 'utf8')
                record_desc = str(record_desc, 'utf8')
                record_quality = str(record_quality, 'utf8')
            yield Reads(seq_id=record_id, sequence=record_seq, desc=record_desc, quality=record_quality)

# K-mer method==========================================================================================================
    def get_k_mer(self, k: int = 21):
        """Get K-mer sequence for each sequence from fastq file."""
        for nucl in self.parse():
            yield from nucl.k_mer(k)

    @staticmethod
    def _stat_k_mer(sequence_obj: Sequence, k: int):
        return sequence_obj.count_k_mer(k)

    def stat_k_mer(self, k: int = 21, num_processing: int = 4):
        """Count the frequency of each k-mer."""
        with tqdm(total=self.num_reads) as pbar:
            params = ((nucl, k) for nucl in self.parse())
            tkm = TaskManager(num_processing=num_processing, params=params)
            rets = tkm.parallel_run_func(func=self._stat_k_mer, call_back_func=lambda _: pbar.update())
            k_mer_count = FuncDict()
            for ret in rets:
                k_mer_count + ret.get()
            k_mer_count.sort_by_keys()
            return k_mer_count

# Other method==========================================================================================================
    @staticmethod
    def generate_random_fastq(
        num_records: int,
        record_length: Union[int, list, tuple] = 150,
        base_bias: Union[float, list] = 1.0
    ) -> Generator[str, Any, None]:
        """
        Generate a random FASTQ file.
        :param num_records: Number of records
        :param record_length: Length or length range of records
        :param base_bias: Base bias
        """
        seed = list('abcdefghijklmnopqrstuvwxyz0123456789'.upper())
        q_value = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~'
        if isinstance(base_bias, float):
            base_bias = [base_bias for _ in range(5)]
        for _ in range(num_records):
            record_id = '@' + ''.join([choice(seed) for _ in range(10)])
            length = randint(min(record_length), max(record_length)) \
                if isinstance(record_length, (list, tuple)) else record_length
            record_seq = ''.join(choices(population="AGCTN", weights=base_bias, k=length))
            Q_value = ''.join(choices(population=q_value, k=length))
            yield f'{record_id}\n{record_seq}\n+\n{Q_value}'

    def sample(self, num_reads: int, other = None):
        rand_int_list = set(choices(range(self.num_reads), k=num_reads))
        while len(rand_int_list) < num_reads:
            rand_int_list.add(randint(0, self.num_reads))
        params = [(self.name, rand_int_list)]
        if other:
            params.append((other.name, rand_int_list))
        tkm = TaskManager(params=params, num_processing=2)
        tkm.parallel_run_func(func=_sample)



"""
File: genotype.py
Description: Instantiate a GT file object.
CreateDate: 2023/10/26
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union, List, Tuple
from typing_extensions import Literal
from functools import partial
from io import TextIOWrapper
from os import getcwd, name, makedirs
from os.path import abspath
from warnings import filterwarnings
from collections import defaultdict
from re import sub
from tqdm import tqdm
from natsort import natsort_key
from numpy import nan
from statsmodels.formula.api import ols
from statsmodels.stats.anova import anova_lm
from pandas import Series, DataFrame, read_table, read_csv, read_excel, concat
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from seaborn import heatmap, clustermap
from click import echo
from pybioinformatic.task_manager import TaskManager
from pybioinformatic.biopandas import read_file_as_df_from_stdin, interval_stat
from pybioinformatic.biomatplotlib import set_custom_font
filterwarnings("ignore")


# =====================================================================================================================#
# NOTICE:                                                                                                              #
# The "__check_hom" and "__get_all_allele" functions cannot be defined in the "GenoType" class.                        #
# Because the "GenoType.parallel_stat_MHM" method runs with multiprocessing,                                           #
# the subprocess cannot call the private method of the "GenoType" class.                                               #
# =====================================================================================================================#
def __check_hom(row: Series) -> Series:
    """Check which samples are homozygous genotypes at specific loci."""
    genotype_set = {'AT', 'TA',
                    'AG', 'GA',
                    'AC', 'CA',
                    'GC', 'CG',
                    'GT', 'TG',
                    'CT', 'TC'}
    data = [str(value) not in genotype_set and '/' not in str(value) and str(value) != ''
            for value in row]
    return Series(data, row.index)


def __get_all_allele(row: Series) -> str:
    """Get all allele at specified loci."""
    all_allele = ''
    ref = row[0]
    for value in row[1:]:
        if '/' in str(value):
            left, right = str(value).split('/')[0], str(value).split('/')[1]
            if 'ins' in left:
                all_allele += 'I'
            elif 'del' in left:
                all_allele += 'D'
            elif left == '-':
                all_allele += ref[0]
            else:
                all_allele += left[0]
            if 'ins' in right:
                all_allele += 'I'
            elif 'del' in right:
                all_allele += 'D'
            elif right == '-':
                all_allele += ref[0]
            else:
                all_allele += right[0]
        elif 'ins' in str(value):
            all_allele += 'II'
        elif 'del' in str(value):
            all_allele += 'DD'
        elif str(value) == '-':
            all_allele += ref[0] * 2
        else:
            value = str(value).replace('*', 'U')
            all_allele += value
    return all_allele


def __allele_sort(allele: str) -> str:
    if '/' in allele:
        return '/'.join(sorted(allele.split('/')))
    elif 'ins' in allele or 'del' in allele:
        return allele
    else:
        return ''.join(sorted(allele))


def allele_sort(df: DataFrame) -> DataFrame:
    snp_ref = df.iloc[:, :4]
    sorted_allele = df.iloc[:, 4:].applymap(__allele_sort, na_action='ignore')
    df = snp_ref.join(sorted_allele)
    return df


def filter_mendelian_errors(df: DataFrame, parents: List[str], offsprings: List[str] = None) -> Tuple[DataFrame, DataFrame]:
    def get_offspring_probable_gt(row: Series) -> Series:
        if len(row) == 6:  # 提供了2个亲本
            row['probable_gt1'] = row.iloc[4][0] + row.iloc[5][0]
            row['probable_gt2'] = row.iloc[4][0] + row.iloc[5][1]
            row['probable_gt3'] = row.iloc[4][1] + row.iloc[5][0]
            row['probable_gt4'] = row.iloc[4][1] + row.iloc[5][1]
        else:  # 仅提供了1个亲本
            row['probable_gt1'] = row.iloc[4][0]
            row['probable_gt2'] = row.iloc[4][1]
        return row

    def filter_snp(row: Series) -> Series:
        num_correct_gt = len([i for i in row.index if 'probable_gt' in i])
        if num_correct_gt == 4:
            correct_gt = list(row[-num_correct_gt:])
            if all([gt in correct_gt for gt in row[4:-num_correct_gt]]):
                return row[:-num_correct_gt]
            else:
                return Series()
        elif num_correct_gt == 2:
            correct_gt = set(row[-num_correct_gt:])
            if all([set(gt) & correct_gt for gt in row[4:-num_correct_gt]]):
                return row[:-num_correct_gt]
            else:
                return Series()
        else:
            return Series()

    # 等位基因排序
    df = allele_sort(df)
    # 拆解亲本和子代基因型并过滤孟德尔错误
    loci_info = df.columns.tolist()[:4]
    samples = df.columns.tolist()[4:]
    if offsprings is None:  # 如果不提供子代样本名
        offsprings = list(set(samples) - set(parents))  # 默认将除了亲本以外的样本都视为子代
    offsprings.sort(key=natsort_key)
    if 0 < len(parents) <= 2:
        parents_gt = allele_sort(df.loc[:, loci_info + parents].apply(get_offspring_probable_gt, axis=1))
        num_probable_gt = len(parents) * 2
        offspring_gt = concat(
            objs=[
                df.loc[:, loci_info + offsprings],
                parents_gt.iloc[:, -num_probable_gt:]
            ],
            axis=1
        ).apply(filter_snp, axis=1).dropna()
    else:
        echo(f"\033[31mError: Invalid number of parents: {len(parents)}.\033[0m", err=True)
        exit()

    pass_df = df.loc[offspring_gt.index].reset_index(drop=True)  # 符合孟德尔遗传的位点
    filtered_df = df.loc[~df.index.isin(offspring_gt.index)].reset_index(drop=True)  # 不符合孟德尔遗传的位点

    return pass_df, filtered_df


def stat_MHM(df: DataFrame) -> DataFrame:
    """
    Calculate MissRate, HetRate, and MAF (MHM) from specified DataFrame.
    The top 4 columns of DataFrame must be SNP ID, chromosome, position, and reference sequence info respectively,
    and the header of DataFrame is not None.
    """
    snp_ref = df.iloc[:, :4]
    # Calculate MissRate
    sample_num = len(df.columns.tolist()) - 4
    df['MissRate(%)'] = df.isnull().sum(axis=1) / sample_num * 100
    # Calculate HetRate
    df.fillna('', inplace=True)  # fill NA
    df['HetRate(%)'] = df.iloc[:, 4:-1].apply(lambda row: (1 - __check_hom(row).sum() / (row != '').sum()) * 100, axis=1)
    # Calculate MAF
    df['all_gt'] = df.iloc[:, 3:-2].apply(lambda row: __get_all_allele(row), axis=1)
    df['total'] = df['all_gt'].apply(len)
    df['A'] = df['all_gt'].str.count('A') / df['total']
    df['G'] = df['all_gt'].str.count('G') / df['total']
    df['C'] = df['all_gt'].str.count('C') / df['total']
    df['T'] = df['all_gt'].str.count('T') / df['total']
    df['D'] = df['all_gt'].str.count('D') / df['total']
    df['I'] = df['all_gt'].str.count('I') / df['total']
    df['U'] = df['all_gt'].str.count('U') / df['total']
    df['MAF'] = df.loc[:, ['A', 'G', 'C', 'T', 'D', 'I', 'U']].apply(lambda row: sorted(row, reverse=True)[1], axis=1)
    # Output results
    df = df.loc[:, ['MissRate(%)', 'HetRate(%)', 'MAF']]
    merge = snp_ref.join(df)
    return merge


def genotype_phenotype_analysis(
    genotype: Series,
    trait: DataFrame,
    trait_name: str,
    mark_sample_list: list = None,
    out_path: str = '.'
) -> Series:
    makedirs(out_path, exist_ok=True)
    site_id = genotype['ID']
    df = genotype[4:].to_frame('Genotype')
    df = concat([df, trait[trait_name]], axis=1)
    df.index.name = 'Sample'

    # 使用statsmodels进行ANOVA
    model = ols(f'{trait_name} ~ C(Genotype)', data=df).fit()
    anova_results = anova_lm(model)
    F = anova_results['F'][0]
    p_value = anova_results['PR(>F)'][0]

    i = 0
    x_tick_labels = []
    for gt, grouped_df in df.groupby('Genotype'):
        df.loc[df.Genotype == gt, 'x'] = i

        plt.boxplot(
            x=grouped_df[trait_name],
            positions=[i],
            medianprops={'color': 'black'}
        )

        plt.scatter(
            x=[i for _ in range(len(grouped_df[trait_name]))],
            y=grouped_df[trait_name]
        )

        x_tick_labels.append(gt)
        i += 1

    if mark_sample_list:
        mark_sample_df = df.loc[mark_sample_list].sort_values(by=trait_name)
        mark_sample_df['Diff'] = mark_sample_df[trait_name].diff()
        mark_sample_df.fillna(0, inplace=True)
        mark_sample_df.loc[mark_sample_df.Diff < 1, 'dy'] = mark_sample_df['Diff'].max() / 10
        mark_sample_df.loc[mark_sample_df.Diff >= 1, 'dy'] = -mark_sample_df['Diff'].max() / 10
        plt.scatter(
            x='x',
            y=trait_name,
            c='r',
            data=mark_sample_df
        )

        for sample in mark_sample_list:
            arrow_x = mark_sample_df.loc[sample]['x']
            arrow_y = mark_sample_df.loc[sample][trait_name]
            dy = mark_sample_df.loc[sample]['dy']
            plt.arrow(
                x=arrow_x,
                y=arrow_y,
                dx=0.3,
                dy=dy,
                linestyle='--',
                linewidth=0.6,
                color='r',
                width=0,
                head_width=0
            )

            plt.text(
                x=arrow_x + 0.31,
                y=arrow_y + dy,
                s=sample,
                fontsize=6,
                color='r'
            )

    plt.xticks(range(len(x_tick_labels)), x_tick_labels)
    title = plt.title(f'site: {site_id}\ntrait: {trait_name}\nF: {F:.3f}  p value: {p_value:.4f}')
    title.set_linespacing(1.5)  # 1.5倍行间距
    plt.savefig(f'{out_path}/{site_id}_{trait_name}.png', dpi=150, bbox_inches='tight')
    plt.savefig(f'{out_path}/{site_id}_{trait_name}.pdf', dpi=150, bbox_inches='tight')
    plt.cla()

    result = Series(
        data=[site_id, trait_name, F, p_value],
        index=['ID', 'Trait', 'F', 'P_value']
    )
    return result


class GenoType:
    """
    Standard genotype file object.
    """

    def __init__(self, path: Union[str, TextIOWrapper]):
        self.name = abspath(path) if isinstance(path, str) else abspath(path.name)
        self.__open = open(path) if isinstance(path, str) else path

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

    def to_dataframe(
        self,
        sheet: Union[str, int, List[Union[str, int]]] = 0,
        index_col: Union[int, str] = None,
        sort_allele: bool = True
    ) -> DataFrame:
        try:  # read from text file
            if 'gz' in self.name:
                df = read_csv(self.name, sep='\t', index_col=index_col)
            else:
                df = read_csv(self.__open, sep='\t', index_col=index_col)
        except UnicodeDecodeError:  # read from Excel file
            df = read_excel(self.name, sheet, index_col=index_col)
        if index_col:
            df.sort_index(key=natsort_key, inplace=True)  # sort by index
        else:
            df.sort_values([df.columns[1], df.columns[2]], key=natsort_key, inplace=True)  # sort by values
        if sort_allele:
            df = allele_sort(df)
        return df

    def to_dataframes(
        self,
        sheet: Union[str, int, List[Union[str, int]]] = 0,
        index_col: Union[int, str] = None,
        chunk_size: int = 10000
    ) -> DataFrame:
        if 'stdin' in self.name:  # read from stdin
            dfs = read_file_as_df_from_stdin(chunk_size=chunk_size, index_col=index_col)
        elif self.name.endswith('gz'):  # read from compressed text (xx.gz) file
            dfs = read_table(self.name, chunksize=chunk_size, index_col=index_col)
        else:
            try:  # read from text file
                dfs = read_table(self.__open, chunksize=chunk_size, index_col=index_col)
            except UnicodeDecodeError:  # read from Excel file
                echo('\033[33mWarning: Excel file cannot be processed in chunks.\033[0m', err=True)
                return self.to_dataframe(sheet, index_col)
        return dfs

    def filter_mendelian_errors(
        self,
        parents: List[str],
        offsprings: List[str] = None,
        num_processing: int = 1
    ) -> Tuple[DataFrame, DataFrame]:
        dfs = self.to_dataframes()
        params = ((df, parents, offsprings) for df in dfs)  # Index of DataFrame is 0, 1, 2, ...
        tkm = TaskManager(num_processing=num_processing, params=params)
        rets = tkm.parallel_run_func(filter_mendelian_errors)
        t = [i.get() for i in rets]  # [(pass_df, filtered_df), (), ...]

        pass_dfs = [i[0] for i in t]
        pass_df = concat(pass_dfs)
        pass_df.sort_values(pass_df.columns.tolist()[1:3], inplace=True, key=natsort_key)
        pass_df.reset_index(drop=True, inplace=True)

        filtered_dfs = [i[1] for i in t]
        filtered_df = concat(filtered_dfs)
        filtered_df.sort_values(filtered_df.columns.tolist()[1:3], inplace=True, key=natsort_key)
        filtered_df.reset_index(drop=True, inplace=True)

        return pass_df, filtered_df

    def filter_allele(self, min_alleles: int, max_alleles: int) -> DataFrame:
        df = self.to_dataframe(index_col=0)
        sub_df = df.iloc[:, 3:]
        combined = (
            sub_df
            .fillna('')  # 将NaN替换为空字符串
            .stack()  # 展开为多级索引Series
            .groupby(level=0)  # 按原始行索引分组
            .agg(''.join)  # 连接所有字符串
        )
        result = (
            combined
            .apply(lambda s: ''.join(sorted(set(s))))  # 去重并排序
            .rename('combined')  # 重命名结果列
        )
        df = concat([df, result], axis=1)
        columns = df.columns.tolist()[:-1]
        filtered_df = df.loc[(df.combined.str.len() <= max_alleles) & (df.combined.str.len() >= min_alleles), columns]
        filtered_df.sort_index(inplace=True)
        return filtered_df

    def parallel_stat_MHM(self, num_processing: int = 1) -> DataFrame:
        """Calculate the MissRate, HetRate and MAF (MHM) of SNP sites from GT files parallely."""
        dfs = self.to_dataframes()
        params = ((df,) for df in dfs)  # Index of DataFrame is 0, 1, 2, ...
        tkm = TaskManager(num_processing=num_processing, params=params)
        ret = tkm.parallel_run_func(stat_MHM)
        stat_dfs = [i.get() for i in ret]
        # Merge results of each multiprocessing
        stat_df = concat(stat_dfs)
        stat_df.sort_values(stat_df.columns.tolist()[1:3], inplace=True, key=natsort_key)
        stat_df.fillna(0, inplace=True)
        return stat_df

    def merge(
        self, other,
        how: Literal['inner', 'outer'],
        sheet1: Union[str, int, List[Union[str, int]]] = None,
        sheet2: Union[str, int, List[Union[str, int]]] = None
    ) -> DataFrame:
        df1 = self.to_dataframe(sheet=sheet1)
        df1[df1.columns[1]] = df1[df1.columns[1]].astype(str)
        df1[df1.columns[2]] = df1[df1.columns[2]].astype(str)
        df1['tmp'] = df1[df1.columns[1]] + '_' + df1[df1.columns[2]] + '_' + df1[df1.columns[3]]
        df1.set_index('tmp', drop=True, inplace=True)
        df1 = df1.iloc[:, 4:]

        df2 = other.to_dataframe(sheet=sheet2)
        df2[df2.columns[1]] = df2[df2.columns[1]].astype(str)
        df2[df2.columns[2]] = df2[df2.columns[2]].astype(str)
        df2['tmp'] = df2[df2.columns[1]] + '_' + df2[df2.columns[2]] + '_' + df2[df2.columns[3]]
        df2.set_index('tmp', drop=True, inplace=True)
        df2 = df2.iloc[:, 4:]

        merge = df1.join(other=df2, how=how)
        samples = merge.columns.tolist()
        samples.sort(key=natsort_key)
        merge = merge.loc[:, samples]
        merge.insert(0, 'Chr', [i.split('_')[0] for i in merge.index.tolist()])
        merge.insert(1, 'Pos', [i.split('_')[1] for i in merge.index.tolist()])
        merge.insert(2, 'Ref', [i.split('_')[2] for i in merge.index.tolist()])
        merge.index.name = 'ID'
        merge.rename(index=lambda i: '_'.join(i.split('_')[:2]), inplace=True)
        merge.fillna('NA', inplace=True)
        merge.sort_index(key=natsort_key, inplace=True)
        return merge

    @staticmethod
    def __draw_gs_heatmap(
        consistency_df: DataFrame,
        cmap: str = "crest",
        output_path: str = getcwd(),
        font_name: str = 'Arial'
    ) -> None:
        # Set font and dpi.
        dpi = 300
        if len(consistency_df) <= 40:
            font_size = 8
        elif 40 < len(consistency_df) <= 100:
            font_size = 6
        elif 100 < len(consistency_df) <= 150:
            font_size = 4
        else:
            font_size = 1
            dpi = 1000
        if name == 'nt':
            plt.rcParams['font.family'] = font_name
        else:
            set_custom_font(font_name)
        plt.rcParams['pdf.fonttype'] = 42
        plt.rcParams['font.size'] = font_size
        # Set figure attributions.
        plt.figure(figsize=(15, 10))
        # Plot heatmap.
        ax = heatmap(
            data=consistency_df,
            cmap=cmap,
            linecolor='w',
            linewidths=0.5,
            xticklabels=True,
            yticklabels=True,
            cbar_kws={'shrink': 0.4}
        )
        plt.tick_params('both', length=0)  # Set scale length.
        # Set color bar ticks and ticks label.
        min_gs = consistency_df.min(numeric_only=True).min() + 1
        max_gs = consistency_df.max(numeric_only=True).max()
        min_value = int(min_gs)
        max_value = int(max_gs)
        step = int((max_value - min_value) / 10)
        step = 1 if step == 0 else step
        cbar = ax.collections[0].colorbar
        cbar_ticks = [i for i in range(min_value, max_value + step, step) if i <= 100]
        cbar.set_ticks(cbar_ticks)
        cbar.set_ticklabels(cbar_ticks, fontsize=8)
        cbar.ax.tick_params(width=0.3)
        # # Save figure.
        plt.savefig(f'{output_path}/Consistency.heatmap.pdf', bbox_inches='tight', dpi=dpi)

    @staticmethod
    def __draw_gs_cluster_heatmap(
        consistency_df: DataFrame,
        cmap: str = "vlag",
        output_path: str = getcwd(),
        font_name: str = 'Arial'
    ) -> None:
        dpi = 300
        if len(consistency_df) <= 40:
            font_size = 8
        elif 40 < len(consistency_df) <= 100:
            font_size = 6
        elif 100 < len(consistency_df) <= 150:
            font_size = 4
        else:
            font_size = 1
            dpi = 1000
        if name == 'nt':
            plt.rcParams['font.family'] = font_name
        else:
            set_custom_font(font_name)
        plt.rcParams['pdf.fonttype'] = 42
        plt.rcParams['font.size'] = font_size
        for i in range(len(consistency_df.columns)):
            consistency_df.iloc[i, i] = 100
            for j in range(i + 1, len(consistency_df.columns)):
                consistency_df.iloc[i, j] = consistency_df.iloc[j, i]
        ax = clustermap(
            data=consistency_df,
            cmap=cmap,
            xticklabels=True    ,
            yticklabels=True,
            dendrogram_ratio=(0.1, 0.1),  # tree diagram percentage
            cbar_pos=(0.999, 0.6, 0.03, 0.3)  # cbar pos
        )
        ax.savefig(f'{output_path}/Consistency.cluster.heatmap.pdf', bbox_inches='tight', dpi=dpi)

    def self_compare(
        self, other,
        sheet1: Union[str, int, List[Union[str, int]]] = None,
        sheet2: Union[str, int, List[Union[str, int]]] = None,
        output_path: str = getcwd()
    ) -> None:
        """Genotype consistency of different test batches in a single sample."""
        # Read in gt.
        df1 = self.to_dataframe(sheet1, index_col=0, sort_allele=False)
        df1[df1.columns[0]] = df1[df1.columns[0]].astype(str)
        df1[df1.columns[1]] = df1[df1.columns[1]].astype(str)
        df1['tmp'] = df1[df1.columns[0]] + '_' + df1[df1.columns[1]] + '_' + df1[df1.columns[2]]
        df1.set_index('tmp', drop=True, inplace=True)
        df2 = other.to_dataframe(sheet2, index_col=0, sort_allele=False)
        df2[df2.columns[0]] = df2[df2.columns[0]].astype(str)
        df2[df2.columns[1]] = df2[df2.columns[1]].astype(str)
        df2['tmp'] = df2[df2.columns[0]] + '_' + df2[df2.columns[1]] + '_' + df2[df2.columns[2]]
        df2.set_index('tmp', drop=True, inplace=True)
        # Check whether the two GT files contain the same loci.
        if df1.index.tolist() != df2.index.tolist():
            echo('\033[31mError: The two GT file loci to be compared are inconsistent.\033[0m', err=True)
            exit()
        # Calculate genotype consistency of each sample under different test batches.
        samples = set(df1.columns[3:]) & set(df2.columns[3:])
        data = []
        with tqdm(total=len(samples), unit='sample') as pbar:
            for sample in samples:
                NA_site = set(df1[sample][df1[sample].isnull()].index) | set(df2[sample][df2[sample].isnull()].index)
                NA_num = len(NA_site)
                series1 = df1[sample][~df1[sample].isnull()]
                series2 = df2[sample][~df2[sample].isnull()]
                IdenticalCount = series1.eq(series2).sum()
                TotalCount = len(df1) - NA_num
                GS = IdenticalCount / TotalCount * 100
                data.append([sample, IdenticalCount, NA_num, TotalCount, GS])
                pbar.update(1)
        bins = range(0, 105, 5)  # Set the consistency statistics interval.
        sample_consistency = DataFrame(data, columns=['SampleName', 'IdenticalCount', 'NaCount', 'TotalCount', 'GS(%)'])
        sample_consistency.sort_values('SampleName', key=natsort_key, inplace=True)
        sample_consistency.fillna(0.000, inplace=True)
        interval_stat_df = interval_stat(ser=sample_consistency['GS(%)'], bins=bins, precision=0, name='Count')
        # Write results to output file.
        sample_consistency.to_csv(f'{output_path}/Sample.consistency.xls', sep='\t', index=False)
        interval_stat_df.to_csv(f'{output_path}/Interval.stat.xls', sep='\t', header=False)

    def compare(
        self, other,
        sheet1: Union[str, int, List[Union[str, int]]] = None,
        sheet2: Union[str, int, List[Union[str, int]]] = None,
        cmap: str = "crest",
        output_path: str = getcwd(),
        font_name: str = 'Arial'
    ) -> None:
        """
        Calculate genotype consistency.
        """
        # Step1: Read GT file as DataFrame.
        df1 = self.to_dataframe(sheet1)  # index = 0, 1, 2, ...
        df2 = other.to_dataframe(sheet2)  # index = 0, 1, 2, ...
        # Step2: Select the site intersection of two GT files.
        left_on = df1.columns[0]
        right_on = df2.columns[0]
        merge = df1.merge(df2, left_on=left_on,  # index = 0, 1, 2, ...
                          right_on=right_on)  # Avoid inconsistency between the two GT file ID fields
        # Step3: Calculate genotype consistency.
        df1_sample_num = len(df1.columns.tolist()) - 4
        left_sample_range = list(range(4, 4 + df1_sample_num))
        if df1.columns[0] == df2.columns[0]:
            right_sample_range = list(range(len(df1.columns) + 3, len(merge.columns)))
        else:
            right_sample_range = list(range(len(df1.columns) + 4, len(merge.columns)))
        consistency_df = DataFrame()
        sample_pair = set()
        data = []
        with tqdm(total=len(left_sample_range), unit='sample') as pbar:  # Show process bar
            for index1 in left_sample_range:
                gt1 = merge.iloc[:, index1]
                gt1_NA = set(gt1[gt1.isnull()].index)
                gt1.name = sub(r'_x\b', '', gt1.name)
                pbar.set_description(f'Processing {gt1.name}')
                for index2 in right_sample_range:
                    gt2 = merge.iloc[:, index2]
                    gt2_NA = set(gt1[gt2.isnull()].index)
                    NA_site_index = gt1_NA | gt2_NA
                    NA_num = len(NA_site_index)
                    TotalCount = len(merge) - NA_num
                    gt2.name = sub(r'_y\b', '', gt2.name)
                    gt1 = gt1[~gt1.isnull()]
                    gt2 = gt2[~gt2.isnull()]
                    IdenticalCount = gt1.eq(gt2).sum()
                    GS = 0.00 if TotalCount == 0 else '%.2f' % (IdenticalCount / TotalCount * 100)
                    data.append([gt1.name, gt2.name, IdenticalCount, NA_num, TotalCount, GS])
                    if gt1.name != gt2.name:
                        if (f'{gt1.name}-{gt2.name}' not in sample_pair) and \
                                (f'{gt2.name}-{gt1.name}' not in sample_pair):
                            consistency_df.loc[gt2.name, gt1.name] = GS
                            sample_pair.add(f'{gt1.name}-{gt2.name}')
                            sample_pair.add(f'{gt2.name}-{gt1.name}')
                    else:
                        consistency_df.loc[gt1.name, gt2.name] = ''
                pbar.update(1)
        header = ['Sample1', 'Sample2', 'IdenticalCount', 'NaCount', 'TotalCount', 'GS(%)']
        fmt1 = DataFrame(data, columns=header)
        fmt1.sort_values(['Sample1', 'GS(%)', 'Sample2'], key=natsort_key, inplace=True, ascending=[True, False, True])
        fmt1.to_csv(f'{output_path}/Sample.consistency.fmt1.xls', sep='\t', index=False)
        # Step4: Draw consistency heatmap.
        consistency_df.to_csv(f'{output_path}/Sample.consistency.fmt2.xls', sep='\t', na_rep='')
        self.__draw_gs_heatmap(
            consistency_df=read_table(f'{output_path}/Sample.consistency.fmt2.xls', index_col=0),
            cmap=cmap,
            output_path=output_path,
            font_name=font_name
        )
        # Step5: Draw consistency cluster heatmap.
        self.__draw_gs_cluster_heatmap(
            consistency_df=read_table(f'{output_path}/Sample.consistency.fmt2.xls', index_col=0),
            cmap=cmap,
            output_path=output_path,
            font_name=font_name
        )
        # Step6: Output GT file of test sample.
        right_sample_range.insert(0, 0)  # Only output site ID
        right_sample_range.insert(1, len(df1.columns) + 2)
        test_sample_gt_df = merge.iloc[:, right_sample_range]
        test_sample_gt_df.rename(columns=lambda i: sub(r'_[xy]\b', '', str(i)), inplace=True)
        test_sample_gt_df.sort_values(merge.columns[0], key=natsort_key, inplace=True)
        test_sample_gt_df.to_csv(f'{output_path}/Compare.GT.xls', sep='\t', index=False, na_rep='NA')

    def find_min_distinguishing_snps(self) -> DataFrame:
        """
        Find the smallest set of SNP loci that can distinguish all samples.
        """
        df = self.to_dataframe(index_col=0, sort_allele=True)
        non_sample_columns = ['Chrom', 'Position', 'Ref']
        sample_columns = [col for col in df.columns if col not in non_sample_columns]
        # 只保留样本列
        genotype_matrix = df[sample_columns].values

        # 初始化等价类（初始所有样本在同一类）
        equivalence_classes = [set(range(len(sample_columns)))]
        selected_snps = []

        # 当存在包含多个样本的等价类时继续
        while any(len(eq) > 1 for eq in equivalence_classes):
            best_snp = None
            best_gain = -1
            best_new_classes = None

            # 遍历所有SNP位点
            for snp_idx in range(len(df)):
                if snp_idx in selected_snps:
                    continue

                # 创建新的等价类划分
                new_classes = []

                # 对每个现有等价类进行细分
                for eq in equivalence_classes:
                    # 按当前SNP的基因型细分
                    subgroups = defaultdict(set)
                    for sample_idx in eq:
                        genotype = genotype_matrix[snp_idx, sample_idx]
                        subgroups[genotype].add(sample_idx)

                    # 添加非空子组
                    for subgroup in subgroups.values():
                        if subgroup:
                            new_classes.append(subgroup)

                # 计算信息增益（减少的未区分样本对数）
                current_undistinguished = sum(len(eq) * (len(eq) - 1) // 2 for eq in equivalence_classes)
                new_undistinguished = sum(len(eq) * (len(eq) - 1) // 2 for eq in new_classes)
                gain = current_undistinguished - new_undistinguished

                # 选择增益最大的SNP
                if gain > best_gain:
                    best_gain = gain
                    best_snp = snp_idx
                    best_new_classes = new_classes

            # 如果没有找到增益的SNP，停止
            if best_gain <= 0:
                break

            # 添加最佳SNP并更新等价类
            selected_snps.append(best_snp)
            equivalence_classes = best_new_classes

        min_snps = df.iloc[selected_snps][non_sample_columns + sample_columns]
        min_snps.sort_index(key=natsort_key, inplace=True)
        return min_snps

    def draw_genotype_heatmap(
        self,
        genotype_colors: List[str] = None,
        font_name: str = 'Arial',
        out_file: str = 'genotype_heatmap.pdf'
    ) -> None:
        if genotype_colors is None:
            genotype_colors = ['#2c7bb6', '#ffffbf', '#d7191c']
        plt.rcParams['pdf.fonttype'] = 42
        if name == 'nt':
            plt.rcParams['font.family'] = font_name
        else:
            set_custom_font(font_name)

        # 1. 读取基因型数据
        df = self.to_dataframe(index_col=0, sort_allele=True)

        # 2. 数据预处理
        # 提取样本列（排除元数据列）
        df.sort_index(key=natsort_key, inplace=True)
        sorted_samples = df.columns.tolist()[3:]
        sorted_samples.sort(key=natsort_key)
        df = df.loc[:, ['Chrom', 'Position', 'Ref'] + sorted_samples]
        sample_columns = [col for col in df.columns if col not in ['Chrom', 'Position', 'Ref']]
        # 创建基因型矩阵
        genotype_matrix = df[sample_columns]

        # 3. 基因型编码（将基因型转换为数值）
        def encode_genotype(gt, ref):
            """将基因型编码为数值"""
            if gt == ref + ref:  # 参考纯合子
                return 0
            elif ref in gt and len(set(gt)) == 2:  # 杂合子
                return 1
            elif ref not in gt:  # 变异纯合子
                return 2
            else:  # 缺失或未知
                return nan

        # 创建编码矩阵
        encoded_matrix = DataFrame(index=genotype_matrix.index, columns=genotype_matrix.columns)

        for idx in genotype_matrix.index:
            ref = df.loc[idx, 'Ref']  # 获取当前SNP的参考等位基因
            for col in genotype_matrix.columns:
                gt = genotype_matrix.loc[idx, col]
                encoded_matrix.loc[idx, col] = encode_genotype(gt, ref)

        encoded_matrix = encoded_matrix.astype(float)
        if len(sample_columns)  > len(df) and len(sample_columns) > 50:
            encoded_matrix = encoded_matrix.T  # 转置矩阵，使样本在Y轴

        # 4. 创建自定义颜色映射
        genotype_cmap = ListedColormap(genotype_colors)

        # 5. 使用Seaborn绘制静态热图并添加基因型文本
        plt.figure(figsize=(10, 8))  # 增大图形尺寸以容纳文本

        # 创建热图
        ax = heatmap(
            encoded_matrix,
            cmap=genotype_cmap,
            cbar_kws={
                'ticks': [-0.5, 0.5, 1.5],
                'boundaries': [-1, 0, 1, 2],
                'shrink': 0.2,  # 缩小比例 (0.0-1.0)
                'aspect': 15  # 颜色条宽高比 (宽度/高度)
            },
            linewidths=0.5,
            linecolor='lightgray',
            vmin=-0.5,
            vmax=2.5,
            annot=False,  # 不使用seaborn的注释，而是自定义
            fmt="",
            square=True
        )

        # 设置颜色条标签
        cbar = ax.collections[0].colorbar
        cbar.set_ticklabels(['Ref Homo', 'Hetero', 'Alt Homo'], fontsize=8)
        # cbar.set_label('Genotype Category', fontsize=12)

        # 添加基因型文本
        if len(sample_columns)  > len(df) and len(sample_columns) > 50:
        # 注意：由于我们转置了矩阵，需要调整索引
            for i, sample in enumerate(encoded_matrix.index):  # i: 样本索引 (Y轴)
                for j, snp in enumerate(encoded_matrix.columns):  # j: SNP索引 (X轴)
                    # 获取原始基因型
                    gt = genotype_matrix.loc[snp, sample]  # 注意索引顺序

                    # 获取当前单元格的编码值
                    cell_val = encoded_matrix.iloc[i, j]

                    # 根据背景色选择文本颜色
                    # 黄色背景用黑色文本，其他用白色文本
                    text_color = 'black' if cell_val == 1 else 'white'

                    # 添加文本
                    plt.text(
                        j + 0.5,  # x位置（单元格中心）
                        i + 0.5,  # y位置（单元格中心）
                        gt,  # 基因型文本
                        ha='center',
                        va='center',
                        color=text_color,
                        fontsize=6
                    )
        else:
            # 添加基因型文本（不转置版本）
            # i: SNP索引 (行/Y轴), j: 样本索引 (列/X轴)
            for i, snp in enumerate(encoded_matrix.index):  # 遍历每个SNP（行）
                for j, sample in enumerate(encoded_matrix.columns):  # 遍历每个样本（列）
                    # 获取原始基因型
                    gt = genotype_matrix.loc[snp, sample]

                    # 获取当前单元格的编码值
                    cell_val = encoded_matrix.iloc[i, j]

                    # 根据背景色选择文本颜色
                    # 黄色背景用黑色文本，其他用白色文本
                    text_color = 'black' if cell_val == 1 else 'white'

                    # 添加文本
                    plt.text(
                        j + 0.5,  # x位置（单元格中心，列位置）
                        i + 0.5,  # y位置（单元格中心，行位置）
                        gt,  # 基因型文本
                        ha='center',
                        va='center',
                        color=text_color,
                        fontsize=6
                    )

        # 添加标题和标签
        if len(sample_columns)  > len(df) and len(sample_columns) > 50:
            plt.xlabel('SNP Positions', fontsize=10, labelpad=15)
            plt.ylabel('Samples', fontsize=10, labelpad=15)
        else:
            plt.xlabel('Samples', fontsize=10, labelpad=15)
            plt.ylabel('SNP Positions', fontsize=10, labelpad=15)

        # 旋转SNP标签以提高可读性
        plt.xticks(rotation=45, ha='right', fontsize=8)
        plt.yticks(fontsize=8)

        plt.tight_layout()  # 优化布局
        plt.savefig(out_file, dpi=300, bbox_inches='tight')

    def filter_multi_group_genotypes(self, groups: List[List[str]]) -> DataFrame:
        """
        Based on the provided sample classification information,
        select the loci where the genotypes within the group are consistent
        but the genotypes between groups are not.
        """
        df = self.to_dataframe(index_col=0, sort_allele=True)
        # 提取所有列并处理缺失值
        all_columns = [col for group in groups for col in group]
        df_clean = df.dropna(subset=all_columns).copy()

        # 检查每个分组内部一致性
        group_conditions = []
        for group in groups:
            group_cond = df_clean[group].nunique(axis=1) == 1
            group_conditions.append(group_cond)

        all_group_cond = concat(group_conditions, axis=1).all(axis=1)

        # 为每组创建代表值列
        rep_cols = []
        for i, group in enumerate(groups):
            rep_col = f'rep_{i}'
            df_clean[rep_col] = df_clean[group[0]]
            rep_cols.append(rep_col)

        # 严格检查组间互异性（两两比较）
        distinct_cond = Series(True, index=df_clean.index)
        n_groups = len(groups)

        for i in range(n_groups):
            for j in range(i + 1, n_groups):
                # 检查每组与其他组是否不同
                distinct_cond = distinct_cond & (df_clean[rep_cols[i]] != df_clean[rep_cols[j]])

        # 组合条件
        result = df_clean[all_group_cond & distinct_cond]
        return result.drop(columns=rep_cols)

    def genotype_phenotype_analysis(
        self,
        trait_file: str,
        trait_name: str,
        font_name: str = 'Arial',
        out_path: str = '.',
        mark_sample_list: List[str] = None
    ) -> None:
        """对单个表型性状进行基因型分组分析，即分析不同基因型间样本表型值差异是否显著"""
        plt.rcParams['pdf.fonttype'] = 42
        plt.rcParams['font.size'] = 8
        if name == 'nt':
            plt.rcParams['font.family'] = font_name
        else:
            set_custom_font(font_name)
        trait_df = read_csv(trait_file, sep='\t').groupby('ID').mean()
        ret = self.to_dataframe(index_col=None, sort_allele=False).apply(
            partial(
                genotype_phenotype_analysis,
                trait=trait_df,
                trait_name=trait_name,
                mark_sample_list=mark_sample_list,
                out_path=out_path
            ),
            axis=1
        )
        ret.to_csv(f'{out_path}/site_sig_stats.xls', sep='\t', na_rep='NA', index=False)



"""
File: biomatplotlib.py
Description: Drawing aid library for matplotlib.
CreateDate: 2024/7/18
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import List, Literal, Union
from os import listdir
from math import trunc
from numpy import random
import matplotlib.axes
import matplotlib.colors as mc
import matplotlib.font_manager as fm
from matplotlib.pyplot import rcParams
from click import echo


def set_custom_font(font_name: str):
    path = '/'.join(__file__.split('/')[:-1]) + '/font'
    all_fonts = [i.split('.')[0] for i in listdir(path)]
    if font_name not in all_fonts:
        msg = (f"\033[33mWarning: {font_name} font not found. Available fonts: {' '.join(all_fonts)}\n"
               f"Try to use the system-provided fonts.\033[0m")
        echo(msg, err=True)
        rcParams['font.family'] = font_name
    else:
        font_path = f'{path}/{font_name}.ttf'
        fm.fontManager.addfont(font_path)
        rcParams['font.family'] = fm.FontProperties(fname=font_path).get_name()


def generate_unique_colors(num_colors: int) -> List[str]:
    """Randomly generates a specified number of non-redundant hexadecimal color codes."""

    def __generate_unique_colors(num_colors):
        colors = set()
        while len(colors) < num_colors:
            r, g, b = random.rand(3)
            color = (r, g, b)
            colors.add(color)
        return colors

    unique_colors = __generate_unique_colors(num_colors)
    hex_colors = [mc.to_hex(color) for color in unique_colors]
    return hex_colors


def format_xticks_by_kmg(
    axes: matplotlib.axes.Axes,
    min_tick: Union[int, float] = None,
    max_tick: Union[int, float] = None,
    tick_unit: Union[int, float] = None,
    decimals: int = 2,
    rotation: Union[int, float] = None,
    grid: bool = True,
    **kwargs
) -> matplotlib.axes.Axes:
    xticks = axes.get_xticks()
    min_x = trunc(float(min(xticks))) if min_tick is None else min_tick
    max_x = trunc(float(max(xticks))) if max_tick is None else max_tick
    yticks = axes.get_yticks()
    min_y = trunc(float(min(yticks)))
    max_y = trunc(float(max(yticks)))
    if max_x / 10 ** 3 <= 1000:
        step = 10 ** 3
        unit = 'Kb'
        step_unit = 10 ** 3
    elif max_x / 10 ** 6 <= 1000:
        step = 10 ** 6
        unit = 'Mb'
        step_unit = 10 ** 6
    else:
        step = 10 ** 9
        unit = 'Gb'
        step_unit = 10 ** 9
    if tick_unit is not None:
        step *= tick_unit
        step = trunc(step)
    x_ticks = [i for i in range(min_x, max_x - step, step)]
    x_ticks.append(max_x)
    s = f'%.{decimals}f'
    x_labels = [s % (i / step_unit) for i in range(min_x, max_x - step, step)]
    x_labels.append(s % (max_x / step_unit))
    axes.set_xticks(x_ticks, x_labels, rotation=rotation)
    axes.set_xlim([min_x, max_x])
    axes.set_xlabel(unit, loc='right')
    if grid:
        for x in x_ticks:
            axes.plot([x, x], [min_y - 1, max_y + 1], **kwargs)
    return axes


def rotate_ax_tick_labels(
    axes: matplotlib.axes.Axes,
    axis: Literal['x', 'y', 'both'],
    rotation: int = 0,
    fontproperties: dict = None
) -> matplotlib.axes.Axes:
    """Rotate coordinate axis tick labels and fix Unicode minus signs without changing the scale of the axes and their labels."""
    axis = axis.lower()
    for ax in (["x", "y"] if axis == "both" else [axis]):
        # Dynamically obtain tags and process Unicode negative signs
        labels = [t.get_text().replace("−", "-") for t in getattr(axes, f"get_{ax}ticklabels")()]
        # Dynamically set the scale and rotation Angle of the coordinate axes
        getattr(axes, f"set_{ax}ticks")(
            [float(label) for label in labels],
            labels=labels,
            rotation=rotation,
            fontproperties=fontproperties
        )
    return axes



"""
File: util.py
Description: Util module.
CreateDate: 2024/5/12
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper
from typing import Callable
from natsort import natsort_key
from yaml import safe_load
from schema import Schema, SchemaError
from click import echo


def check_config(module_dict:dict, yaml_file: TextIOWrapper):
    echo('\033[36mCheck params.\033[0m', err=True)
    config_schema = Schema(module_dict)

    config = safe_load(yaml_file)
    try:
        config_schema.validate(config)
        echo("\033[32mConfiguration verification passed.\n\033[0m", err=True)
    except SchemaError as e:
        echo(f"\033[31mConfig error: {e}\033[0m", err=True)
        exit()
    else:
        return config


class FuncDict(dict):
    def sort_by_keys(self, key: Callable = natsort_key,  reverse: bool = False):
        sorted_dict = FuncDict({key: self[key] for key in sorted(self, key=key, reverse=reverse)})
        self.clear()
        self.update(sorted_dict)

    def __getitem__(self, key):
        """Returns the same value as the key when a non-existent key is accessed."""
        try:
            return super().__getitem__(key)
        except KeyError:
            return key

    def __add__(self, other):
        for k, v in other.items():
            if k in self:
                self[k] += v
            else:
                self[k] = v



"""
File: biomysql.py
Description: MySQL python API.
CreateDate: 2024/10/15
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from sshtunnel import SSHTunnelForwarder
from pymysql import connect


class BioMySQL:
    def __init__(self,
                 mysql_host: str,
                 mysql_port: int,
                 mysql_user: str,
                 mysql_password: str,
                 ssh_ip: str = None,
                 ssh_port: int = None,
                 ssh_user: str = None,
                 ssh_password: str = None,
                 remote_bind_ip: str = None,
                 remote_bind_port: int = None,
                 local_bind_ip: str = None,
                 local_bind_port: int = None):
        # MySQL params
        self.mysql_host = mysql_host if not local_bind_ip else local_bind_ip
        self.mysql_port = mysql_port if not local_bind_port else local_bind_port
        self.mysql_user = mysql_user
        self.mysql_password = mysql_password
        # SSH params
        self.ssh_ip = ssh_ip
        self.ssh_port = ssh_port
        self.ssh_user = ssh_user
        self.ssh_password = ssh_password
        self.remote_bind_ip = remote_bind_ip
        self.remote_bind_port = remote_bind_port
        self.local_bind_ip = local_bind_ip
        self.local_bind_port = local_bind_port

    def __enter__(self):
        if self.local_bind_ip:
            self.connect_server_by_ssh()
            self.ssh_server.start()
        return self.connect_mysql()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.mysql_server.cursor().close()
            self.mysql_server.close()
            self.ssh_server.stop()
        except Exception:
            pass

    def connect_server_by_ssh(self):
        self.ssh_server = SSHTunnelForwarder(
            (self.ssh_ip, self.ssh_port),
            ssh_username=self.ssh_user,
            ssh_password=self.ssh_password,
            remote_bind_address=(self.remote_bind_ip, self.remote_bind_port),
            local_bind_address=(self.local_bind_ip, self.local_bind_port)
        )
        return self.ssh_server

    def connect_mysql(self):
        conf = {
            'host': self.mysql_host,
            'port': self.mysql_port,
            'user': self.mysql_user,
            'password': self.mysql_password,
            'charset': 'utf8',
        }
        self.mysql_server = connect(**conf)
        return self.mysql_server



#!/usr/bin/env python
"""
File: base.py
Description: Next-generation sequencing (NGS) data analysis module.
CreateDate: 2025/4/18
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Dict, List
from os import makedirs
from os.path import abspath, exists
from re import sub
from shutil import which
from click import echo
from pybioinformatic.task_manager import TaskManager
from pybioinformatic.util import FuncDict


def check_cmds(cmds_list: List[str]):
    """
    Check if the required commands are available in the system's PATH.

    This function iterates through a list of command names, checking if each
    command is present in the system's PATH using the `which` function. If a
    command is found, its path is printed to standard error, and True is appended
    to the internal tracking list. If a command is not found, an error message is
    printed, and False is appended to the list. If any command is missing, the
    program will terminate.

    Parameters:
    cmds_list (List[str]): A list of command names to check for availability.

    Returns:
    None

    Raises:
    SystemExit: If any of the commands in the list are not found in the system's
                PATH, the program will terminate.
    """
    flag = []
    echo('\033[36mCheck dependency.\033[0m', err=True)
    for software in cmds_list:
        path = which(software)
        if path:
            echo(f'{software}: {path}', err=True)
            flag.append(True)
        else:
            echo(f'{software}: command not found', err=True)
            flag.append(False)
    if not all(flag):
        exit()


def check_R_packages(packages_list: List[str]):
    """
    Check if the specified R packages are installed on the system.

    This function iterates through a list of R package names and verifies their
    installation by executing an R command. If a package is not installed, an error
    message is displayed, and the program terminates if any package is missing.
    """
    flag = []
    tkm = TaskManager(num_processing=1)
    for package in packages_list:
        cmd = '''R CMD Rscript -e 'if(requireNamespace("%s", quietly = TRUE)) {print("True")} else {print("False")}' ''' % package
        stdout, _ = tkm.echo_and_exec_cmd(cmd=cmd, show_cmd=False, pipe=True)
        if 'True' in stdout:
            flag.append(True)
        else:
            flag.append(False)
            echo(f'\033[31mR package {package} has not been installed.\033[0m', err=True)
    if not all(flag):
        exit()


def parse_sample_info(sample_info: str) -> Dict[str, list]:
    """
    Parse sample information from a file and return it as a dictionary.

    This function reads a tab-separated file containing sample information and
    constructs a dictionary where each key is a sample name, and the value is a
    list of associated data such as read paths and comparative combinations.
    The function validates the input file for common issues like missing fields
    or invalid file paths.

    Parameters
    ----------
    sample_info : str
        Path to the file containing sample information. The file should be
        tab-separated with at least three columns: sample name, read1 path,
        and read2 path.
        For RNA-seq, the fourth column is sample comparative combinations.
        For ChIP-seq, the first and second columns are read1 and read2 path
        of ChIP, the fourth column is reference genome path, and the fifth
        and sixth columns are read1 and read2 path of Input.

    Returns
    -------
    Dict[str, list]
        A dictionary where keys are sample names (str) and values are lists
        containing read paths and additional comparative combinations.
    """
    sample_info_dict: Dict[str, List[str]] = {}
    # for RNA-seq:
    # {
    #   sample_name: [read1_path, read2_path, comparative_combination, ...],
    #   ...
    # }
    # for ChIP-seq:
    # {
    #   sample_name: [read1_path_of_ChIP, read2_path_of_ChIP, ref_genome, read1_path_of_Input, read2_path_of_Input,...],
    #   ...
    # }
    with open(sample_info) as f:
        line_num = 0
        for line in f:
            line_num += 1
            if not line.startswith('#'):
                split = line.strip().split('\t')
                if len(split) < 3:
                    echo(f'\033[31mError: Invalid sample info at line {line_num}.\033[0m', err=True)
                elif not exists(split[1]):
                    echo(f'\033[31mError: Invalid read1 path at line {line_num}.\033[0m', err=True)
                elif not exists(split[2]):
                    echo(f'\033[31mError: Invalid read2 path at line {line_num}.\033[0m', err=True)
                else:
                    sample_info_dict[split[0]] = split[1:]
    return sample_info_dict


def build_ref_index(
    fasta_file: str,
    bwa_exe: str = None,
    gatk_exe: str = None,
    samtools_exe: str = None,
    hisat2_build_exe: str = None,
    bowtie_build: str = None,
    bowtie2_build: str = None,
    large: bool = False
) -> str:
    software_list = [
        which(i)
        for i in [bwa_exe, gatk_exe, samtools_exe, hisat2_build_exe, bowtie_build, bowtie2_build]
        if i is not None and which(i)
    ]
    if not software_list:
        msg = '\033[33mWaring: No any valid executable file was specified, ignore building any index of fasta file.\033[0m'
        echo(msg, err=True)
        exit()
    fasta_file = abspath(fasta_file)
    cmd_dict = {
        'bwa': f' index {fasta_file} {fasta_file}',
        'gatk': f' CreateSequenceDictionary -R {fasta_file}',
        'samtools': f' faidx {fasta_file}',
        'hisat2-build': f' {fasta_file} {fasta_file}',
        'bowtie-build': f' {fasta_file} {fasta_file}',
        'bowtie2-build': f' {fasta_file} {fasta_file}',
    }
    try:
        cmd = '\n'.join(
            [
                i + cmd_dict[i.split('/')[-1]]
                for i in software_list
            ]
        )
    except KeyError as s:
        echo(f'\033[31mError: Invalid command {s}.\033[0m', err=True)
        exit()
    else:
        if large:
            cmd = sub('build ', 'build --large-index ', cmd)
    return cmd


class NGSAnalyser:
    """
    \033[32mNext-generation sequencing (NGS) data analysis pipeline.\033[0m

    \033[34m:param read1: Pair end read1 fastq file.\033[0m
    \033[34m:param read2: Pair end read2 fastq file.\033[0m
    \033[34m:param ref_genome: Reference genome fasta file.\033[0m
    \033[34m:param output_path: Output path.\033[0m
    \033[34m:param num_threads: Number of threads.\033[0m
    \033[34m:param sample_name: Sample name.\033[0m
    \033[34m:param exe_path_dict: Software executable file path dict.\033[0m
    """

    _exe_set = {'fastp', 'bwa', 'samtools', 'hisat2', 'gatk', 'bowtie', 'bowtie2', 'samblaster'}

    def __init__(
        self,
        read1: str,
        read2: str,
        ref_genome: str,
        output_path: str,
        num_threads: int = 10,
        sample_name: str = None,
        exe_path_dict: Dict[str, str] = None
    ):
        if exe_path_dict is None:
            exe_path_dict = {}
        ## Input file
        self.read1 = abspath(read1)
        self.read2 = abspath(read2)
        self.ref_genome = abspath(ref_genome)
        self.output_path = abspath(output_path)
        self.num_threads = num_threads
        self.sample_name = self.read1.split('/')[-1].split('.')[0] if sample_name is None else sample_name
        self.exe_path_dict = FuncDict(
            {
                k: which(v)
                for k, v in exe_path_dict.items()
                if k in self._exe_set and which(v) is not None
            }
        )
        ## Key output file
        # QC results file
        self.qc_path = f'{self.output_path}/01.QC/{self.sample_name}'
        self.read1_clean = f'{self.qc_path}/{self.sample_name}_1_clean.fq.gz'
        self.read2_clean = f'{self.qc_path}/{self.sample_name}_2_clean.fq.gz'
        # bwa mem results file
        self.mapping_path = f'{self.output_path}/02.mapping/{self.sample_name}'
        self.bwa_mem_raw_bam = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.bam'
        self.bwa_mem_map30_bam = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.map30.bam'
        self.bwa_mem_map30_markdup_bam = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.map30.markdup.bam'
        self.bwa_mem_map30_markdup_bam_metrics = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.map30.markdup.bam.metrics'
        # depth file
        self.depth = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.map30.markdup.bam.depth'
        # bowtie2 reasults file
        self.bowtie2_raw_bam = f'{self.mapping_path}/{self.sample_name}.bt2.sort.bam'
        self.bowtie2_filtered_bam = f'{self.mapping_path}/{self.sample_name}.bt2.sort.filtered.bam'
        # hisat2 results file
        self.hisat2_raw_bam = f'{self.mapping_path}/{self.sample_name}.ht2.sort.bam'

    @staticmethod
    def other_options(cmd: str, other_options: dict) -> str:
        other_options = ' '.join([f'{k} {v}' for k, v in other_options.items()])
        cmd += f' {other_options}'
        return cmd

    def run_fastp(
        self,
        read1_raw: str = None,
        read2_raw: str = None,
        read1_clean: str = None,
        read2_clean: str = None,
        **other_options
    ) -> str:
        fastp = which(self.exe_path_dict['fastp'])
        read1_raw = abspath(read1_raw) if read1_raw else self.read1
        read2_raw = abspath(read2_raw) if read2_raw else self.read2
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        makedirs(self.qc_path, exist_ok=True)
        cmd = (
            f'{fastp} '
            f'-w {self.num_threads} '
            f'-i {read1_raw} '
            f'-I {read2_raw} '
            f'-o {read1_clean} '
            f'-O {read2_clean}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_hisat2(
        self,
        read1_clean: str = None,
        read2_clean: str = None,
        out_bam: str = None,
        **other_options
    ) -> str:
        hisat2 = which(self.exe_path_dict['hisat2'])
        samtools = which(self.exe_path_dict['samtools'])
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        out_bam = abspath(out_bam) if out_bam else self.hisat2_raw_bam
        makedirs(self.mapping_path, exist_ok=True)
        hisat2_cmd = (
            f'{hisat2} -p {self.num_threads} '
            f'-x {self.ref_genome} '
            f'-1 {read1_clean} '
            f'-2 {read2_clean} '
            f'--summary-file {self.mapping_path}/{self.sample_name}.ht2.log'
        )
        hisat2_cmd = self.other_options(hisat2_cmd, other_options) if other_options else hisat2_cmd
        cmd = f'{hisat2_cmd} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} - -o {out_bam}'
        return cmd

    def run_bwa_mem(
        self,
        read1_clean: str = None,
        read2_clean: str = None,
        out_bam: str = None,
        markdup: bool = False,
        **other_options
    ) -> str:
        bwa = which(self.exe_path_dict['bwa'])
        samblaster = which(self.exe_path_dict['samblaster'])
        samtools = which(self.exe_path_dict['samtools'])
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        out_bam = abspath(out_bam) if out_bam else self.bwa_mem_raw_bam
        makedirs(self.mapping_path, exist_ok=True)
        bwa_cmd = (
            fr'{bwa} mem -t {self.num_threads} '
            fr'-R "@RG\tID:{self.sample_name}\tSM:{self.sample_name}\tPL:ILLUMINA" '
            fr'{self.ref_genome} {read1_clean} {read2_clean}'
        )
        bwa_cmd = self.other_options(bwa_cmd, other_options) if other_options else bwa_cmd
        if markdup:
            cmd = f'{bwa_cmd} | {samblaster} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} -o {out_bam}'
        else:
            cmd = f'{bwa_cmd} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} -o {out_bam}'
        return cmd

    def filter_reads_by_mapQ(
        self,
        bam_file: str = None,
        out_bam: str = None,
        map_q: int = 30
    ) -> str:
        if map_q not in range(1, 60):
            msg = f'\033[31mError: Invalid mapping Q value: "{map_q}". It must be an integer between 0 and 60.\033[0m'
            echo(message=msg, err=True)
            exit()
        samtools = which(self.exe_path_dict['samtools'])
        makedirs(self.mapping_path, exist_ok=True)
        bam_file = abspath(bam_file) if bam_file else self.bwa_mem_raw_bam
        out_bam = abspath(out_bam) if out_bam else self.bwa_mem_map30_bam
        awk = r'''awk '{if($1~/@/){print}else{if( $7 == "=" &&  $5 >= %s ){print $0}}}' ''' % map_q
        cmd = f'{samtools} view -h {bam_file} | {awk}| samtools view -bS -T {self.ref_genome} - -o {out_bam}'
        return cmd

    def mark_duplicates(
        self,
        bam_file: str = None,
        out_bam: str = None,
        out_metrics: str = None,
        **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        samtools = which(self.exe_path_dict['samtools'])
        makedirs(self.mapping_path, exist_ok=True)
        bam_file = abspath(bam_file) if bam_file else self.bwa_mem_map30_bam
        out_bam = abspath(out_bam) if out_bam else self.bwa_mem_map30_markdup_bam
        out_metrics = abspath(out_metrics) if out_metrics else self.bwa_mem_map30_markdup_bam_metrics
        cmd = (
            f'{gatk} MarkDuplicates '
            f'-I {bam_file} '
            f'-M {out_metrics} '
            f'-O {out_bam}'
        )
        cmd = self.other_options(cmd, other_options) if other_options else cmd
        cmd += f'\n{samtools} index {out_bam}'
        return cmd

    def stats_depth(
        self,
        bam_file: str = None,
        out_depth: str = None,
        **other_options
    ) -> str:
        samtools = which(self.exe_path_dict['samtools'])
        makedirs(self.mapping_path, exist_ok=True)
        bam_file = abspath(bam_file) if bam_file else self.bwa_mem_map30_markdup_bam
        out_depth = abspath(out_depth) if out_depth else self.depth
        cmd = f'{samtools} depth {bam_file} -o {self.mapping_path}/{out_depth}'
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_bowtie2(
        self,
        read1_clean: str = None,
        read2_clean: str = None,
        markdup: bool = False,
        out_bam: str = None,
        **other_options
    ) -> str:
        bowtie2 = which(self.exe_path_dict['bowtie2'])
        samblaster = which(self.exe_path_dict['samblaster'])
        samtools = which(self.exe_path_dict['samtools'])
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        out_bam = abspath(out_bam) if out_bam else self.bowtie2_raw_bam
        makedirs(self.mapping_path, exist_ok=True)
        bowtie2_cmd = (
            f'{bowtie2} '
            f'-p {self.num_threads} '
            f'-x {self.ref_genome} '
            f'-1 {read1_clean} '
            f'-2 {read2_clean}'
        )
        bowtie2_cmd = self.other_options(bowtie2_cmd, other_options) if other_options else bowtie2_cmd
        if markdup:
            cmd = f'{bowtie2_cmd} | {samblaster} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} -o {out_bam}'
        else:
            cmd = f'{bowtie2_cmd} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} -o {out_bam}'
        return cmd

    def filter_reads_by_samtools(
        self,
        bam_file: str = None,
        out_bam: str = None,
        **other_options
    ) -> str:
        samtools = which(self.exe_path_dict['samtools'])
        bam_file = abspath(bam_file) if bam_file else self.bowtie2_raw_bam
        out_bam = abspath(out_bam) if out_bam else self.bowtie2_filtered_bam
        cmd = f'{samtools} view {bam_file} -o {out_bam}'
        cmd = self.other_options(cmd, other_options) if other_options else cmd
        return cmd



"""
File: gtf.py
Description: Instantiate a GTF file object.
CreateDate: 2021/12/1
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper, StringIO
from typing import Dict, List, Union, Generator, Any, Iterable
from os.path import abspath
from gzip import GzipFile
from pandas import read_table, Series, DataFrame
from natsort import natsorted
import matplotlib
import matplotlib.pyplot as plt
from click import open_file, Choice
from pybioinformatic.fasta import Fasta
from pybioinformatic.sequence import Nucleotide
from pybioinformatic.biomatplotlib import set_custom_font


class Gtf:
    def __init__(self, path: Union[str, TextIOWrapper]):
        if isinstance(path, str):
            self.name = abspath(path)
            if not path.endswith('gz'):
                self.__open = open(path)
                self.line_num = sum(1 for line in self.__open if not line.startswith('#'))
                self.__open.seek(0)
                self.anno_line_num = sum(1 for line in self.__open if line.startswith('#'))
                self.__open.seek(0)
            else:
                self.__open = GzipFile(self.name, 'rb')
                self.line_num = sum(1 for line in self.__open if not str(line, 'utf8').startswith('#'))
                self.__open.seek(0)
                self.anno_line_num = sum(1 for line in self.__open if str(line, 'utf8').startswith('#'))
                self.__open.seek(0)
        else:
            self.name = abspath(path.name)
            if path.name == '<stdin>' or '/dev' in path.name:
                string_io = StringIO()
                for line in path:
                    string_io.write(line)
                string_io.seek(0)
                self.__open = string_io
                self.line_num = sum(1 for line in self.__open if not line.startswith('#'))
                self.__open.seek(0)
                self.anno_line_num = sum(1 for line in self.__open if line.startswith('#'))
                self.__open.seek(0)
            else:
                if not self.name.endswith('gz'):
                    self.__open = path
                    self.line_num = sum(1 for line in self.__open if not line.startswith('#'))
                    self.__open.seek(0)
                    self.anno_line_num = sum(1 for line in self.__open if line.startswith('#'))
                    self.__open.seek(0)
                else:
                    self.__open = GzipFile(self.name, 'rb')
                    self.line_num = sum(1 for line in self.__open if not str(line, 'utf8').startswith('#'))
                    self.__open.seek(0)
                    self.anno_line_num = sum(1 for line in self.__open if str(line, 'utf8').startswith('#'))
                    self.__open.seek(0)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def parse(self):
        """Parse information of each column of GTF file line by line."""
        for line in self.__open:
            line = str(line, 'utf8') if self.name.endswith('gz') else line
            if not line.startswith('#') and line.strip():
                split = line.strip().split('\t')
                chr_num, source, feature = split[0], split[1], split[2]
                start, end, score, strand, frame = split[3], split[4], split[5], split[6], split[7]
                attr_list = [attr for attr in split[8].replace('"', '').split(';') if attr]
                attr_dict = {attr.strip().split(' ')[0]: attr.strip().split(' ')[1]
                             for attr in attr_list if attr}
                yield chr_num, source, feature, start, end, score, strand, frame, attr_dict

    def to_dataframe(self):
        def parse_attribute(row: Series):
            attr_list = row['Attribute'].split('; ')
            for _attr in attr_list:
                k, v = _attr.split(' ')
                v = v.replace('"', '').replace(';', '')
                row[k] = v
            return row

        names = ['Chromosome', 'Source', 'Feature', 'Start', 'End', 'Score', 'Strand', 'Frame', 'Attribute']
        if self.name.endswith('gz'):
            df = read_table(self.name, header=None, names=names, dtype=str, comment='#')
        else:
            df = read_table(self.__open, header=None, names=names, dtype=str, comment='#')
        df[['Start', 'End']] = df[['Start', 'End']].astype(int)
        df = df.apply(func=parse_attribute, axis=1)
        df.drop(labels='Attribute', inplace=True, axis=1)
        raw = set(names[:-1])
        attr = list(set(df.columns) - raw)
        attr.sort()
        columns = names[:-1]
        columns.extend(attr)
        df = df.loc[:, columns]
        self.__open.seek(0)
        return df

    def merge_by_transcript(self) -> DataFrame:
        def apply_func(row: DataFrame):
            chr_name = row['Chromosome'].unique()[0]
            strand = row['Strand'].unique()[0]
            start, end = row['Start'].min(), row['End'].max()
            data = [chr_name, start, end, strand]
            index = ['Chromosome', 'Start', 'End', 'Strand']
            ser = Series(data=data, index=index)
            return ser

        df = self.to_dataframe()
        merge_df = df.groupby('transcript_id').apply(func=apply_func)  # transcript_id Chromosome Start End Strand
        return merge_df

    def get_exon_dict(self) -> Dict[str, List[Dict[str, Union[int, str]]]]:
        """Save all exons' information in the GTF file into the dictionary."""
        # exon_dict = {
        #              Chr_num: [{transcript_id: str, start: int, end: int, strand: str}, {}, ...],
        #              Chr_num: [{exon}, {exon}, ...], ...
        #              }
        exon_dict = {}
        for line in self.parse():
            if line[2] == 'exon':
                item = {
                    'id': line[8]['transcript_id'],
                    'start': int(line[3]),
                    'end': int(line[4]),
                    'strand': line[6]
                }
                if line[0] in exon_dict:
                    exon_dict[line[0]].append(item)
                else:
                    exon_dict[line[0]] = [item]
        return exon_dict

    def get_non_redundant_exon(self) -> Dict[str, List[Dict[str, Union[int, str]]]]:
        """Get non-redundant exon of each gene."""
        # non_redundant_exon_dict = {
        #                            Chr_num: [{gene_id: str, start: int, end: int, strand: str}, {}, ...],
        #                            Chr_num: [{exon}, {exon}, ...], ...
        #                            }
        non_redundant_exon_dict = {}
        gene_id = ''
        raw_exon_list = []
        line_count = 0
        for line in self.parse():
            line_count += 1
            if line[0] not in non_redundant_exon_dict:
                non_redundant_exon_dict[line[0]] = []
            if line[2] == 'exon':
                item = {'id': line[8]['gene_id'], 'start': int(line[3]), 'end': int(line[4]), 'strand': line[6]}
                if gene_id:
                    if gene_id == line[8]['gene_id']:
                        raw_exon_list.append(item)
                    else:
                        i = 0
                        while i + 1 < len(raw_exon_list):
                            raw_exon_list.sort(key=lambda d: (d['start'], d['end']))
                            if raw_exon_list[i]['end'] >= raw_exon_list[i + 1]['start']:
                                raw_exon_list[i + 1]['start'] = raw_exon_list[i]['start']
                                del raw_exon_list[i]
                                raw_exon_list.sort(key=lambda d: (d['start'], d['end']))
                                i = 0
                            elif raw_exon_list[i + 1]['end'] <= raw_exon_list[i]['end']:
                                del raw_exon_list[i + 1]
                                raw_exon_list.sort(key=lambda d: (d['start'], d['end']))
                                i = 0
                            else:
                                i += 1
                        non_redundant_exon_dict[line[0]].extend(raw_exon_list)
                        gene_id = line[8]['gene_id']
                        raw_exon_list = [item]
                else:
                    gene_id = line[8]['gene_id']
                    raw_exon_list.append(item)
                if line_count == self.line_num:
                    i = 0
                    while i + 1 < len(raw_exon_list):
                        raw_exon_list.sort(key=lambda d: (d['start'], d['end']))
                        if raw_exon_list[i]['end'] >= raw_exon_list[i + 1]['start']:
                            raw_exon_list[i + 1]['start'] = raw_exon_list[i]['start']
                            del raw_exon_list[i]
                            raw_exon_list.sort(key=lambda d: (d['start'], d['end']))
                            i = 0
                        elif raw_exon_list[i + 1]['end'] <= raw_exon_list[i]['end']:
                            del raw_exon_list[i + 1]
                            raw_exon_list.sort(key=lambda d: (d['start'], d['end']))
                            i = 0
                        else:
                            i += 1
                    non_redundant_exon_dict[line[0]].extend(raw_exon_list)
        return non_redundant_exon_dict

    def get_gene_dict(self) -> Dict[str, List[Dict[str, Union[int, str]]]]:
        """Get gene dict. The start and end based on gene length, not based on chromosome length."""
        gene_dict = {}  # {gene_id: [{start: int, end: int, strand: str}, {}, ...], ...}
        non_redundant_exon_dict = self.get_non_redundant_exon()
        for chr_num, l in non_redundant_exon_dict.items():
            gene_id = gene_start = None
            raw_exon_list = []
            l.sort(key=lambda i: (
                i['id'], i['start'], i['end']))  # [{gene_id: str, start: int, end: int, strand: str}, ...]
            for exon_dict in l:  # {gene_id: str, start: int, end: int, strand: str}
                if gene_id:
                    if exon_dict['id'] == gene_id:
                        item = {'start': exon_dict['start'] - gene_start, 'end': exon_dict['end'] - gene_start,
                                'strand': exon_dict['strand']}
                        raw_exon_list.append(item)
                        if exon_dict == l[-1]:
                            if raw_exon_list[0]['strand'] == '-':
                                raw_exon_list.sort(key=lambda i: -i['end'])
                                gene_length = raw_exon_list[0]['end']
                                for d in raw_exon_list:
                                    start = gene_length - d['end']
                                    end = gene_length - d['start']
                                    d['start'] = start
                                    d['end'] = end
                            gene_dict[gene_id] = raw_exon_list
                    else:
                        if raw_exon_list[0]['strand'] == '-':
                            raw_exon_list.sort(key=lambda i: -i['end'])
                            gene_length = raw_exon_list[0]['end']
                            for d in raw_exon_list:
                                start = gene_length - d['end']
                                end = gene_length - d['start']
                                d['start'] = start
                                d['end'] = end
                        gene_dict[gene_id] = raw_exon_list
                        gene_id = exon_dict['id']
                        gene_start = exon_dict['start']
                        item = {'start': exon_dict['start'] - gene_start, 'end': exon_dict['end'] - gene_start,
                                'strand': exon_dict['strand']}
                        raw_exon_list = [item]
                        if exon_dict == l[-1]:
                            gene_dict[gene_id] = raw_exon_list
                else:
                    gene_id = exon_dict['id']
                    gene_start = exon_dict['start']
                    item = {'start': exon_dict['start'] - gene_start, 'end': exon_dict['end'] - gene_start,
                            'strand': exon_dict['strand']}
                    raw_exon_list.append(item)
                    if exon_dict == l[-1]:
                        if raw_exon_list[0]['strand'] == '-':
                            raw_exon_list.sort(key=lambda i: -i['end'])
                            gene_length = raw_exon_list[0]['end']
                            for d in raw_exon_list:
                                start = gene_length - d['end']
                                end = gene_length - d['start']
                                d['start'] = start
                                d['end'] = end
                        gene_dict[gene_id] = raw_exon_list
        return gene_dict

# GTF file Natural sorting method=======================================================================================
    @staticmethod
    def __sort(line: str):
        split = line.strip().split('\t')
        chromosome = split[0]
        start, end = int(split[3]), int(split[4])
        attr_list = split[8].split(' ')
        gene_id = attr_list[attr_list.index('gene_id') + 1]
        transcript_id = attr_list[attr_list.index('transcript_id') + 1]
        return chromosome, gene_id, transcript_id, start, end

    def sort(self) -> str:
        """Sort naturally in sequence according to chromosome, gene id, transcript id, start, and end."""
        if self.name.endswith('gz'):
            l = [
                str(line, 'utf8').strip()
                for line in self.__open
                if not str(line, 'utf8').startswith('#') and str(line, 'utf8').strip()
            ]
        else:
            l = [
                line.strip()
                for line in self.__open
                if not line.startswith('#') and line.strip()
            ]
        sorted_gtf = natsorted(l, key=self.__sort)
        if not isinstance(self.__open, list):
            self.__open.seek(0)
        return '\n'.join(sorted_gtf)

# Sequence extraction method============================================================================================
    def get_cDNA(self, fasta_file: Union[str, TextIOWrapper]) -> Generator[Nucleotide, Any, None]:  # return Nucleotide objet generator
        """Extract cDNA sequence from GTF file according large reference sequence."""
        exon_dict = self.get_exon_dict()  # {chr_num: [{tid: str, start: int, end: int, strand: str}, {}, ...], ...}
        for nucl_obj in Fasta(fasta_file).parse():
            cDNA_id = None
            cDNA_seq = []
            cDNA_strand = None
            try:
                exon_list = exon_dict[nucl_obj.id]  # [{id: str, start: int, end: int, strand: str}, {}, ...]
                exon_list.sort(key=lambda item: (item['id'], item['start'], item['end']))
            except KeyError:
                pass  # Some sequences (eg. scaffold, contig) may not have annotation
            else:
                for exon in exon_list:  # {id: str, start: int, end: int, strand: str}
                    exon_seq: str = nucl_obj.seq[exon['start'] - 1:exon['end']]
                    if cDNA_id and cDNA_id == exon['id']:
                        cDNA_seq.append(exon_seq)
                        if exon == exon_list[-1]:
                            cDNA_nucl_obj = Nucleotide(cDNA_id, ''.join(cDNA_seq))
                            if cDNA_strand == '-':
                                cDNA_nucl_obj = -cDNA_nucl_obj
                            cDNA_nucl_obj.id = cDNA_id
                            yield cDNA_nucl_obj
                    elif cDNA_id and cDNA_id != exon['id']:
                        cDNA_nucl_obj = Nucleotide(cDNA_id, ''.join(cDNA_seq))
                        if cDNA_strand == '-':
                            cDNA_nucl_obj = -cDNA_nucl_obj
                        cDNA_nucl_obj.id = cDNA_id
                        yield cDNA_nucl_obj
                        cDNA_seq = [exon_seq]
                        cDNA_id = exon['id']
                        cDNA_strand = exon['strand']
                        if exon == exon_list[-1]:
                            cDNA_nucl_obj = Nucleotide(cDNA_id, ''.join(cDNA_seq))
                            if cDNA_strand == '-':
                                cDNA_nucl_obj = -cDNA_nucl_obj
                            cDNA_nucl_obj.id = cDNA_id
                            yield cDNA_nucl_obj
                    elif not cDNA_id:
                        cDNA_seq = [exon_seq]
                        cDNA_id = exon['id']
                        cDNA_strand = exon['strand']
                        if exon == exon_list[-1]:
                            cDNA_nucl_obj = Nucleotide(cDNA_id, ''.join(cDNA_seq))
                            if cDNA_strand == '-':
                                cDNA_nucl_obj = -cDNA_nucl_obj
                            cDNA_nucl_obj.id = cDNA_id
                            yield cDNA_nucl_obj

# File format conversion method=========================================================================================
    def to_bed(self, feature_type: str = 'exon') -> Generator[str, None, None]:
        """Convert the file format from GTF to BED."""
        for line in self.parse():
            if line[2] == feature_type != 'gene':
                yield f"{line[0]}\t{int(line[3]) - 1}\t{line[4]}\t{line[8]['transcript_id']}\t{line[7]}\t{line[6]}"
            elif line[2] == feature_type == 'gene':
                yield f"{line[0]}\t{int(line[3]) - 1}\t{line[4]}\t{line[8]['gene_id']}\t{line[7]}\t{line[6]}"

    def to_gsds(self, feature_type: Choice(['gene', 'transcript']) = 'transcript') -> Generator[str, Any, None]:
        """Convert the file format from GTF to GSDS."""
        if feature_type == 'gene':
            exon_dict = self.get_non_redundant_exon()
        else:
            exon_dict = self.get_exon_dict()
        parent_id = parent_start = None
        for _, l in exon_dict.items():
            l.sort(key=lambda d: (d['id'], d['start'], d['end']))
            for exon in l:
                if not parent_id:
                    parent_id = exon['id']
                    parent_start = exon['start']
                    yield f"{exon['id']}\t{exon['start'] - parent_start}\t{exon['end'] - parent_start}\texon\t."
                else:
                    if parent_id == exon['id']:
                        yield f"{exon['id']}\t{exon['start'] - parent_start}\t{exon['end'] - parent_start}\texon\t."
                    else:
                        parent_id = exon['id']
                        parent_start = exon['start']
                        yield f"{exon['id']}\t{exon['start'] - parent_start}\t{exon['end'] - parent_start}\texon\t."

# plot method===========================================================================================================
    def plot_gene_structure(
        self,
        gene_set: Iterable[str] = None,
        exon_color: str = 'salmon',
        font: str = 'Arial'
    ) -> matplotlib.axes.Axes:
        gene_dict = self.get_gene_dict()  # {gene_id: [{start: int, end: int, strand: str}, {}, ...], ...}
        gene_set = set(gene_set) if gene_set is not None else set(gene_dict.keys())
        # set font size and figsize
        set_custom_font(font)
        plt.rcParams['pdf.fonttype'] = 42
        if len(gene_dict) <= 10:
            figsize = (10.0, 2.0)
            plt.rcParams['font.size'] = 8
        elif 10 < len(gene_dict) <= 30:
            figsize = (10.0, 8.0)
            plt.rcParams['font.size'] = 8
        elif 30 < len(gene_dict) <= 50:
            figsize = (10.0, 8.0)
            plt.rcParams['font.size'] = 6
        elif 50 < len(gene_dict) <= 100:
            figsize = (10.0, 8.0)
            plt.rcParams['font.size'] = 3
        else:
            figsize = (10.0, 8.0)
            plt.rcParams['font.size'] = 1

        # create figure and axes
        fig = plt.figure(figsize=figsize, dpi=300)
        ax = fig.add_subplot(111)

        # plot intron and exon
        y = 1
        label_set = set()
        intron_linewidth = 0.4 if len(gene_dict) <= 100 else 0.05
        for gene_id, exons in gene_dict.items():
            if gene_id in gene_set:
                gene_len = max([i['end'] for i in exons]) - 10
                if 'intron' not in label_set:
                    label = 'intron'
                    label_set.add(label)
                else:
                    label = None
                plt.plot(
                    [0, gene_len],
                    [y, y],
                    color='black',
                    label=label,
                    linewidth=intron_linewidth,
                    zorder=0
                )
                exons.sort(key=lambda i: i['start'])
                for exon in exons:
                    left = exon['start']
                    width = exon['end'] - exon['start'] + 1
                    if 'exon' not in label_set:
                        label = 'exon'
                        label_set.add(label)
                    else:
                        label = None
                    ax.barh(
                        y,
                        width=width,
                        height=0.8,
                        left=left,
                        label=label,
                        color=exon_color,
                        zorder=1
                    )
                y += 1

        ax.set_yticks(range(1, y), gene_dict.keys())
        plt.tick_params(axis='y', length=0)
        ax.spines['top'].set_color(None)
        ax.spines['right'].set_color(None)
        ax.spines['left'].set_color(None)
        plt.legend()

        return ax



#!/usr/bin/env python
"""
File: decompressing_file.py
Description: Decompressing file.
CreateDate: 2022/4/22
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from gzip import GzipFile


def ungz(gz_file):
    """Decompressing gz file."""
    with open(gz_file.replace('.gz', ''), 'ab') as out:
        for line in GzipFile(gz_file):
            out.write(line)
    return gz_file.replace('.gz', '')



#!/usr/bin/env python
"""
File: task_manager.py
Description: Instance a TaskManager.
CreateDate: 2023/9/8
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union, Tuple, Iterable, Callable, Generator
from os import getcwd
from subprocess import run, PIPE
from datetime import datetime
from getpass import getuser
from socket import gethostname
import multiprocessing
from click import echo


class TaskManager:
    def __init__(
        self,
        commands: Iterable[str] = None,
        num_processing: int = None,
        params: Iterable[tuple] = None
    ):
        try:
            self.task = list(commands)
        except TypeError:
            self.task = []
        self.params = params
        self.num_processing = num_processing

    def add_task(self, command: Union[str, Iterable[str]]):
        if isinstance(command, str):
            self.task.append(command)
        else:
            command = list(command)
            self.task.extend(command)

    def del_task(self, index: Union[int, str] = None):
        if isinstance(index, int):
            del self.task[index]
        elif isinstance(index, str):
            self.task.remove(index)
        else:
            self.task.pop()

    def clear_task(self):
        self.task = []

    @staticmethod
    def echo_and_exec_cmd(
        cmd: str,
        show_cmd: bool = True,
        pipe: bool = True
    ) -> Union[Tuple[str, str], None]:
        if show_cmd:
            echo(
                f'\033[33m[{getuser()}@{gethostname()}: '
                f'{datetime.now().replace(microsecond=0)} {getcwd()}]\n$ '
                f'\033[0m\033[36m{cmd}\033[0m', err=True
            )
        if pipe:
            ret = run(cmd, shell=True, executable="/bin/bash", stdout=PIPE, stderr=PIPE)
            stdout, stderr = ret.stdout.decode('utf8'), ret.stderr.decode('utf8')
            return stdout, stderr
        else:
            run(cmd, shell=True, executable="/bin/bash")
            return None

    def serial_run_cmd(self, show_cmd: bool = True, pipe: bool = False) -> Generator[Tuple[str, str], None, None]:
        for cmd in self.task:
            yield self.echo_and_exec_cmd(cmd=cmd, show_cmd=show_cmd, pipe=pipe)

    def parallel_run_cmd(self, show_cmd: bool = True, pipe: bool = False) -> list:
        if not self.task:
            echo('\033[31mError: TaskManager has no task.\033[0m', err=True)
            exit()
        results = []
        with multiprocessing.Pool(self.num_processing) as pool:
            for cmd in self.task:
                ret = pool.apply_async(self.echo_and_exec_cmd, args=(cmd, show_cmd, pipe))
                results.append(ret)
            pool.close()
            pool.join()
        return results

    def parallel_run_func(self, func: Callable, call_back_func: Callable = None):
        results = []
        with multiprocessing.Pool(self.num_processing) as pool:
            for param in self.params:
                ret = pool.apply_async(func, args=param, callback=call_back_func)
                results.append(ret)
            pool.close()
            pool.join()
        return results


if __name__ == '__main__':
    tkm = TaskManager()
    cmds = [f'echo Hello World x {i}' for i in range(10)]
    tkm.add_task(cmds)
    tkm.serial_run_cmd()



"""
File: DNA_seq.py
Description: DNA-seq analysis pipeline module.
CreateDate: 2025/4/18
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Dict
from os import makedirs
from os.path import abspath
from shutil import which
from pybioinformatic.NGS.base import NGSAnalyser
from pybioinformatic.util import FuncDict
from pybioinformatic.fasta import Fasta


class GatkSNPCalling(NGSAnalyser):
    """
    Variant calling pipeline using GATK tools.

    This class implements a complete pipeline for SNP calling using GATK's HaplotypeCaller, GenotypeGVCFs,
    and VariantFiltration tools. It handles the generation of GVCF and VCF files, applies filtering to
    variants, and integrates with other NGS analysis steps such as read alignment and quality control.

    Parameters:
        read1 (str): Path to the first paired-end FASTQ file.
        read2 (str): Path to the second paired-end FASTQ file.
        ref_genome (str): Path to the reference genome FASTA file.
        output_path (str): Directory where output files will be stored.
        num_threads (int): Number of threads to use for parallel processing. Default is 10.
        sample_name (str): Name of the sample being analyzed. If not provided, it will be inferred.
        exe_path_dict (Dict[str, str]): Dictionary mapping tool names to their executable paths.
    """
    def __init__(
        self,
        read1: str,
        read2: str,
        ref_genome: str,
        output_path: str,
        num_threads: int = 10,
        sample_name: str = None,
        exe_path_dict: Dict[str, str] = None
    ):
        super().__init__(read1, read2, ref_genome, output_path, num_threads, sample_name, exe_path_dict)
        self.variant_path = f'{self.output_path}/03.variant/{self.sample_name}'
        self.gvcf = f"{self.variant_path}/{self.bwa_mem_map30_markdup_bam.split('/')[-1]}.gvcf.gz"
        self.vcf = f"{self.variant_path}/{self.bwa_mem_map30_markdup_bam.split('/')[-1]}.vcf.gz"
        self.filtered_vcf = f"{self.variant_path}/{self.bwa_mem_map30_markdup_bam.split('/')[-1]}.filtered.vcf.gz"

    def run_HaplotypeCaller(
        self,
        bam_file: str = None,
        out_gvcf: str = None,
        **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        makedirs(self.variant_path, exist_ok=True)
        bam_file = abspath(bam_file) if bam_file else self.bwa_mem_map30_markdup_bam
        out_gvcf = abspath(out_gvcf) if out_gvcf else self.gvcf
        cmd = (f'{gatk} HaplotypeCaller '
               f'-ERC GVCF '
               f'-I {bam_file} '
               f'-R {self.ref_genome} '
               f'-O {out_gvcf}')
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_GenotypeGVCFs(
        self,
        gvcf_file: str = None,
        out_vcf: str = None,
        **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        makedirs(self.variant_path, exist_ok=True)
        gvcf_file = abspath(gvcf_file) if gvcf_file else self.gvcf
        out_vcf = abspath(out_vcf) if out_vcf else self.vcf
        cmd = (
            f'{gatk} GenotypeGVCFs '
            f'-R {self.ref_genome} '
            f'-V {gvcf_file} '
            f'-O {out_vcf}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_VariantFiltration(
        self,
        filter_expression: str = 'QD < 2.0 || MQ < 40.0 || FS > 60.0 || SOR > 3.0',
        vcf_file: str = None,
        out_vcf: str = None,
        **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        makedirs(self.variant_path, exist_ok=True)
        vcf_file = abspath(vcf_file) if vcf_file else self.vcf
        out_vcf = abspath(out_vcf) if out_vcf else self.filtered_vcf
        cmd = (
            f'{gatk} VariantFiltration '
            f'--filter-name  "HARD_TO_VALIDATE" '
            f'--filter-expression "{filter_expression}" '
            f'-R {self.ref_genome} '
            f'-V {vcf_file} '
            f'-O {out_vcf}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def pipeline(
        self,
        fastp_options: Dict[str, str] = None,
        bwa_mem_options: Dict[str, str] = None,
        map_q: int = 30
    ) -> str:
        if fastp_options is None:
            fastp_options = {
                    '-j': f'{self.qc_path}/{self.sample_name}.fastp.json',
                    '-h': f'{self.qc_path}/{self.sample_name}.fastp.html'
                }
        if bwa_mem_options is None:
            bwa_mem_options = {}
        cmds = (
            f'{self.run_fastp(**fastp_options)}\n\n'
            f'{self.run_bwa_mem(**bwa_mem_options)}\n\n'
            f'{self.filter_reads_by_mapQ(bam_file=None, out_bam=None, map_q=map_q)}\n\n'
            f'{self.mark_duplicates()}\n\n'
            f'{self.stats_depth()}\n\n'
            f'{self.run_HaplotypeCaller()}\n\n'
            f'{self.run_GenotypeGVCFs()}\n\n'
            f'{self.run_VariantFiltration()}'
        )
        return cmds


class Macs2PeakCalling:
    """
    Manages the peak calling process using MACS2 for ChIP-Seq data analysis.
    """
    _exe_set = {'macs2'}

    def __init__(
        self,
        ChIP_read1: str,
        ChIP_read2: str,
        Input_read1: str,
        Input_read2: str,
        ref_genome: str,
        output_path: str,
        num_threads: int = 10,
        sample_name: str = None,
        exe_path_dict: Dict[str, str] = None
    ):
        self.ref_genome = abspath(ref_genome)
        self.sample_name = sample_name
        self.output_path = abspath(output_path)
        self.num_threads = num_threads
        if exe_path_dict is None:
            exe_path_dict = {}
        self.exe_path_dict = FuncDict(
            {
                k: which(v)
                for k, v in exe_path_dict.items()
                if k in self._exe_set and which(v) is not None
            }
        )

        self.ChIP_NGSAnalyser = NGSAnalyser(
            read1=abspath(ChIP_read1),
            read2=abspath(ChIP_read2),
            ref_genome=abspath(ref_genome),
            output_path=output_path,
            num_threads=num_threads,
            sample_name=sample_name,
            exe_path_dict=exe_path_dict
        )

        self.Input_NGSAnalyser = NGSAnalyser(
            read1=abspath(Input_read1),
            read2=abspath(Input_read2),
            ref_genome=abspath(ref_genome),
            output_path=output_path,
            num_threads=num_threads,
            sample_name=sample_name,
            exe_path_dict=exe_path_dict
        )

        self.qc_path = f'{self.output_path}/01.QC/{self.sample_name}'
        self.mapping_path = f'{self.output_path}/02.mapping/{self.sample_name}'
        self.peaks_path = f'{self.output_path}/03.peaks/{self.sample_name}'

        self.ChIP_read1_clean = f'{self.qc_path}/{self.sample_name}_ChIP_1_clean.fq.gz'
        self.ChIP_read2_clean = f'{self.qc_path}/{self.sample_name}_ChIP_2_clean.fq.gz'
        self.ChIP_fastp_json = f'{self.qc_path}/{self.sample_name}_ChIP.fastp.json'
        self.ChIP_fastp_html = f'{self.qc_path}/{self.sample_name}_ChIP.fastp.html'

        self.Input_read1_clean = f'{self.qc_path}/{self.sample_name}_Input_1_clean.fq.gz'
        self.Input_read2_clean = f'{self.qc_path}/{self.sample_name}_Input_2_clean.fq.gz'
        self.Input_fastp_json = f'{self.qc_path}/{self.sample_name}_Input.fastp.json'
        self.Input_fastp_html = f'{self.qc_path}/{self.sample_name}_Input.fastp.html'

        self.ChIP_bowtie2_markdup_bam = f'{self.mapping_path}/{self.sample_name}_ChIP.bt2.markdup.sort.bam'
        self.ChIP_bowtie2_markdup_filtered_bam = f'{self.mapping_path}/{self.sample_name}_ChIP.bt2.markdup.sort.filtered.bam'

        self.Input_bowtie2_markdup_bam = f'{self.mapping_path}/{self.sample_name}_Input.bt2.markdup.sort.bam'
        self.Input_bowtie2_markdup_filtered_bam = f'{self.mapping_path}/{self.sample_name}_Input.bt2.markdup.sort.filtered.bam'

    @staticmethod
    def other_options(cmd: str, other_options: dict):
        other_options = ' '.join([f'{k} {v}' for k, v in other_options.items()])
        cmd += f' {other_options}'
        return cmd

    def run_fastp(self, **other_options) -> str:
        ChIP_qc_other_options = {
                '-j': self.ChIP_fastp_json,
                '-h': self.ChIP_fastp_html
            }
        if other_options:
            ChIP_qc_other_options.update(other_options)
        ChIP_qc_cmd = self.ChIP_NGSAnalyser.run_fastp(
            read1_clean=self.ChIP_read1_clean,
            read2_clean=self.ChIP_read2_clean,
            **ChIP_qc_other_options
        )

        Input_qc_other_options = {
                '-j': self.Input_fastp_json,
                '-h': self.Input_fastp_html
            }
        if other_options:
            Input_qc_other_options.update(other_options)
        Input_qc_cmd = self.Input_NGSAnalyser.run_fastp(
            read1_clean=self.Input_read1_clean,
            read2_clean=self.Input_read2_clean,
            **Input_qc_other_options
        )

        cmd = f'{ChIP_qc_cmd}\n{Input_qc_cmd}'
        return cmd

    def run_bowtie2(self, **other_options) -> str:
        ChIP_mapping_cmd = self.ChIP_NGSAnalyser.run_bowtie2(
            read1_clean=self.ChIP_read1_clean,
            read2_clean=self.ChIP_read2_clean,
            markdup=True,
            out_bam=self.ChIP_bowtie2_markdup_bam,
            **other_options
        )
        Input_mapping_cmd = self.Input_NGSAnalyser.run_bowtie2(
            read1_clean=self.Input_read1_clean,
            read2_clean=self.Input_read2_clean,
            markdup=True,
            out_bam=self.Input_bowtie2_markdup_bam,
            **other_options
        )
        cmd = f'{ChIP_mapping_cmd}\n{Input_mapping_cmd}'
        return cmd

    def filter_reads(self, **other_options) -> str:
        ChIP_filter_cmd = self.ChIP_NGSAnalyser.filter_reads_by_samtools(
            bam_file=self.ChIP_bowtie2_markdup_bam,
            out_bam=self.ChIP_bowtie2_markdup_filtered_bam,
            **other_options
        )
        Input_filter_cmd = self.Input_NGSAnalyser.filter_reads_by_samtools(
            bam_file=self.Input_bowtie2_markdup_bam,
            out_bam=self.Input_bowtie2_markdup_filtered_bam,
            **other_options
        )
        cmd = f'{ChIP_filter_cmd}\n{Input_filter_cmd}'
        return cmd

    def run_macs2(
        self,
        genome_size: int = None,
        ChIP_bam: str = None,
        Input_bam: str = None,
        **other_options
    ) -> str:
        macs2 = which(self.exe_path_dict['macs2'])
        makedirs(self.peaks_path, exist_ok=True)
        ChIP_bam = abspath(ChIP_bam) if ChIP_bam else self.ChIP_bowtie2_markdup_filtered_bam
        Input_bam = abspath(Input_bam) if Input_bam else self.Input_bowtie2_markdup_filtered_bam
        if genome_size is None:
            with Fasta(self.ref_genome) as fa:
                genome_size = fa.get_size()
        cmd = (
            f'{macs2} callpeak '
            f'-t {ChIP_bam} '
            f'-c {Input_bam} '
            f'-g {genome_size} '
            f'-n {self.sample_name} '
            f'--outdir {self.peaks_path}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def pipeline(
        self,
        fastp_options: Dict[str, str] = None,
        bowtie2_options: Dict[str, str] = None,
        samtools_filter_reads_options: Dict[str, str] = None,
        macs2_callpeak_options: Dict[str, str] = None
    ) -> str:
        if fastp_options is None:
            fastp_options = {}

        if bowtie2_options is None:
            bowtie2_options = {
                '--very-sensitive': '',
                '--no-mixed': '',
                '--no-discordant': '',
                '-k': '10',
                '-tq': '',
                '-X': '1000',
                '-L': '25'
            }

        if samtools_filter_reads_options is None:
            samtools_filter_reads_options = {
                '-bF': '1804',
                '-f': '2',
                '-q': '20',
                '-e': "'[NM] <= 2'"
            }

        if macs2_callpeak_options is None:
            macs2_callpeak_options = {}

        cmds = (
            f'{self.run_fastp(**fastp_options)}\n\n'
            f'{self.run_bowtie2(**bowtie2_options)}\n\n'
            f'{self.filter_reads(**samtools_filter_reads_options)}\n\n'
            f'{self.run_macs2(**macs2_callpeak_options)}'
        )

        return cmds



"""
File: blast.py
Description: Instantiate a blast format 6 result file object.
CreateDate: 2021/12/4
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper
from typing import Union, Tuple
from os.path import abspath
from pandas import read_table
from click import open_file


class Blast:
    def __init__(self, path: Union[str, TextIOWrapper]):
        if isinstance(path, str):
            self.name = abspath(path)
            self.__open = open(path)
            self.line_num = sum(1 for _ in self.__open)
            self.__open.seek(0)
        else:
            self.name = abspath(path.name)
            if path.name == '<stdin>':
                self.__open = open_file('-').readlines()
                self.line_num = sum(1 for _ in self.__open)
            else:
                self.__open = path
                self.line_num = sum(1 for _ in self.__open)
                self.__open.seek(0)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def parse(self) -> Tuple[str]:
        """Parse information of each column of BLAST format 6 result file line by line."""
        for line in self.__open:
            split = line.strip().split('\t')
            query_id, sbject_id, align_rate = split[0], split[1], split[2]
            align_len, mismatch, gap = split[3], split[4], split[5]
            query_start, query_end = split[6], split[7]
            sbject_start, sbject_end = split[8], split[9]
            e_value, score = split[10], split[11]
            yield query_id, sbject_id, align_rate, align_len, mismatch, gap, \
                query_start, query_end, sbject_start, sbject_end, e_value, score

    def to_dataframe(self):
        names = [
            'Query', 'Sbject', 'AlignRate', 'AlignLength', 'Missmatch', 'GapOpen',
            'QueryStart', 'QueryEnd', 'SbjectStart', 'SbjectEnd', 'Evalue', 'Score'
        ]
        dtype = {
            'Query': str, 'Sbject': str, 'AlignRate': float, 'AlignLength': int, 'Missmatch': int, 'GapOpen': int,
            'QueryStart': int, 'QueryEnd': int, 'SbjectStart': int, 'SbjectEnd': int, 'Evalue': float, 'Score': float
        }
        df = read_table(self.__open, header=None, names=names, dtype=dtype)
        return df

    def get_pair_dict(self, top: int = 3):
        pair_dict = {}  # {query1: {sbject1: [], sbject2: [], ...}, query2: {}, ...}
        for line in self.parse():
            if line[0] not in pair_dict:
                pair_dict[line[0]] = {line[1]: '\t'.join(line[2:])}
            else:
                if len(pair_dict[line[0]]) < top:
                    pair_dict[line[0]][line[1]] = '\t'.join(line[2:])
        return pair_dict

# File format conversion method=========================================================================================
    def to_bed(self, query_is_ref: bool = False) -> str:
        """Transform the result of align the query sequence with the reference sequence into a BED file."""
        content = ''
        for line in self.parse():
            if not query_is_ref:
                if int(line[8]) - int(line[9]) > 0:
                    content += f'{line[1]}\t{int(line[9]) - 1}\t{line[8]}\t{line[0]}\t.\t-\n'
                else:
                    content += f'{line[1]}\t{int(line[8]) - 1}\t{line[9]}\t{line[0]}\t.\t+\n'
            else:
                if int(line[6]) - int(line[7]) > 0:
                    content += f'{line[0]}\t{int(line[7]) - 1}\t{line[6]}\t{line[1]}\t.\t-\n'
                else:
                    content += f'{line[0]}\t{int(line[6]) - 1}\t{line[7]}\t{line[1]}\t.\t+\n'
        return content



from pybioinformatic.bed import Bed
from pybioinformatic.biomysql import BioMySQL
from pybioinformatic.blast import Blast
from pybioinformatic.decompressing_file import ungz
from pybioinformatic.fasta import Fasta
from pybioinformatic.fastq import Fastq
from pybioinformatic.genotype import GenoType
from pybioinformatic.gff import Gff
from pybioinformatic.gtf import Gtf
from pybioinformatic.sequence import Sequence, Nucleotide, Protein, Reads
from pybioinformatic.show_info import Displayer
from pybioinformatic.task_manager import TaskManager
from pybioinformatic.timer import Timer
from pybioinformatic.util import FuncDict, check_config
from pybioinformatic.vcf import VCF

from pybioinformatic.NGS import (
    check_cmds,
    check_R_packages,
    parse_sample_info,
    build_ref_index,
    GatkSNPCalling,
    Macs2PeakCalling,
    RNASeqAnalysisWithReference,
    LncRNAPredictor,
    LncRNATargetPredictor,
    LncRNAClassification
)

from pybioinformatic.biopandas import (
    display_set,
    read_file_as_df_from_stdin,
    merge_duplicate_indexes,
    filter_by_min_value,
    get_FPKM,
    get_TPM,
    dfs_to_excel,
    dataframe_to_str,
    interval_stat
)

from pybioinformatic.biomatplotlib import (
    set_custom_font,
    generate_unique_colors,
    format_xticks_by_kmg,
    rotate_ax_tick_labels
)

__version__ = '1.2.5'
__all__ = [
    'Bed',
    'Blast',
    'BioMySQL',
    'check_cmds',
    'check_R_packages',
    'parse_sample_info',
    'build_ref_index',
    'GatkSNPCalling',
    'Macs2PeakCalling',
    'RNASeqAnalysisWithReference',
    'LncRNAPredictor',
    'LncRNATargetPredictor',
    'LncRNAClassification',
    'ungz',
    'Fasta',
    'Fastq',
    'GenoType',
    'Gff',
    'Gtf',
    'Sequence',
    'Nucleotide',
    'Protein',
    'Reads',
    'Displayer',
    'Timer',
    'TaskManager',
    'VCF',
    'FuncDict',
    'check_config',
    'display_set',
    'read_file_as_df_from_stdin',
    'merge_duplicate_indexes',
    'filter_by_min_value',
    'get_TPM',
    'get_FPKM',
    'dfs_to_excel',
    'dataframe_to_str',
    'interval_stat',
    'generate_unique_colors',
    'set_custom_font',
    'format_xticks_by_kmg',
    'rotate_ax_tick_labels'
]



"""
File: vcf.py
Description: Instantiate a VCF file object.
CreateDate: 2023/10/28
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union, Any, Generator
from io import TextIOWrapper
from os.path import abspath
from gzip import GzipFile
from pandas import read_table


class VCF:
    def __init__(self, path: Union[str, TextIOWrapper]):
        self.name = abspath(path) if isinstance(path, str) else abspath(path.name)
        if isinstance(path, str):
            if path.endswith('gz'):
                self.__open = GzipFile(path)
            else:
                self.__open = open(path)
        else:
            if path.name.endswith('gz'):
                self.__open = GzipFile(path.name)
            else:
                self.__open = path

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

    def __seek_zero(self):
        try:
            self.__open.seek(0)
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def get_header(self):
        for line in self.__open:
            if not isinstance(line, str):
                line = str(line, 'utf8')
            if not line.startswith('##') and line.startswith('#'):
                header = line.strip().replace('#', '').split('\t')
                self.__seek_zero()
                return header

    def to_dataframe(self):
        df = read_table(
            self.__open,
            comment='#',
            header=None,
            names=self.get_header()
        )
        return df

# Format conversion method==============================================================================================
    def to_genotype(self) -> Generator[str, Any, None]:
        """Convert VCF to GenoType."""
        for line in self.__open:
            if not isinstance(line, str):
                line = str(line, 'utf8')
            if not line.startswith('#') and line.strip():
                split = line.strip().split('\t')
                chr_name, position, ID, ref, alts, = split[:5]
                alts = alts.split(',')
                alts.insert(0, ref)
                ID = f'{chr_name}_{position}'
                samples = split[9:]
                samples_gt = []
                for sample in samples:
                    gt = sample.split(':')[0]
                    if gt == './.' or gt == '.|.' or gt == '.':
                        gt = 'NA'
                    else:
                        alleles_index = [int(i) for i in gt.replace('|', '/').split('/')]
                        alt1, alt2 = alts[alleles_index[0]], alts[alleles_index[1]]
                        if len(ref) == len(alt1) == len(alt2) == 1:
                            gt = alt1 + alt2
                        elif len(ref) == len(alt1) == len(alt2) > 1:
                            if ref == alt1 == alt2:
                                gt = '-'
                            else:
                                gt = alt1[0] + alt2[0]
                        elif len(alt1) == len(alt2) > 1:
                            gt = alt1[0] + alt2[0]
                        else:
                            if ref == alt1:
                                alt1 = '-'
                            elif len(ref) == len(alt1):
                                alt1 = alt1[0]
                            elif alt1 != '*':
                                alt1 = f"ins{alt1.replace(ref, '', 1)}" \
                                    if len(ref) < len(alt1) else \
                                    f"del{ref.replace(alt1, '', 1)}"
                            if ref == alt2:
                                alt2 = '-'
                            elif len(ref) == len(alt2):
                                alt2 = alt2[0]
                            elif alt2 != '*':
                                alt2 = f"ins{alt2.replace(ref, '', 1)}" \
                                    if len(ref) < len(alt2) else \
                                    f"del{ref.replace(alt2, '', 1)}"
                            if alt1 == alt2 and ('del' in alt1 or 'ins' in alt1):
                                gt = alt1
                            else:
                                gt = f'{alt1}/{alt2}' if alt1 != alt2 else '-'
                    samples_gt.append(gt)
                samples_gt = '\t'.join(samples_gt)
                yield f'{ID}\t{chr_name}\t{position}\t{ref}\t{samples_gt}'
            elif not line.startswith('##') and line.startswith('#'):
                split = line.strip().split('\t')
                header = 'ID\tChrom\tPosition\tRef\t'
                samples = '\t'.join(split[9:])
                yield header + samples
